"""Utility methods related to Input/Output."""
from pathlib import Path


def is_equal_or_child_of(source1: Path, source2: Path) -> bool:
    """
    Checking if a source equal or a child of another source.

    :param source1: Child source to check
    :type source1: Path
    :param source2: Parent source to check
    :type source2: Path
    :return: Is sources1 relative to source2
    :rtype: bool
    """
    source1, source2 = source1.absolute(), source2.absolute()
    if source2 == source1:
        return True
    return source2 in source1.parents



"""Evaluation of commands map."""
import datetime
import json
from dataclasses import dataclass, field
from pathlib import Path
from typing import Any, Dict, ItemsView, Iterator, KeysView, List, ValuesView

from statue.command import CommandEvaluation
from statue.commands_map import CommandsMap
from statue.constants import DATETIME_FORMAT, ENCODING


@dataclass
class SourceEvaluation:
    """Evaluation result of a source."""

    commands_evaluations: List[CommandEvaluation] = field(default_factory=list)
    source_execution_duration: float = field(default=0)

    def __len__(self) -> int:
        """
        Number of command evaluations.

        :return: command evaluations count
        :rtype: int
        """
        return len(self.commands_evaluations)

    def append(self, command_evaluation: CommandEvaluation) -> None:
        """
        Append new command evaluation.

        :param command_evaluation: Evaluation to be appended
        :type command_evaluation: CommandEvaluation
        """
        self.commands_evaluations.append(command_evaluation)

    def as_dict(self) -> Dict[str, Any]:
        """
        Return source evaluation as json dictionary.

        :return: Self as dictionary
        :rtype: Dict[str, Any]
        """
        return dict(
            commands_evaluations=[
                command_evaluation.as_dict()
                for command_evaluation in self.commands_evaluations
            ],
            source_execution_duration=self.source_execution_duration,
        )

    @property
    def success(self) -> bool:
        """
        All commands evaluations are successful.

        :return: Success statue
        :rtype: bool
        """
        return all(
            commands_evaluation.success
            for commands_evaluation in self.commands_evaluations
        )

    @property
    def commands_number(self):
        """
        Number of commands that were evaluated.

        :return: Counted commands
        :rtype: int
        """
        return len(self.commands_evaluations)

    @property
    def successful_commands_number(self):
        """
        Number of successful commands that were evaluated.

        :return: Counted successful commands
        :rtype: int
        """
        return len(
            [
                commands_evaluation
                for commands_evaluation in self.commands_evaluations
                if commands_evaluation.success
            ]
        )

    @property
    def failed_commands_number(self):
        """
        Number of failed commands that were evaluated.

        :return: Counted failed commands
        :rtype: int
        """
        return self.commands_number - self.successful_commands_number

    @classmethod
    def from_dict(cls, source_evaluation: Dict[str, Any]) -> "SourceEvaluation":
        """
        Read source evaluation from json list.

        :param source_evaluation: Json commands evaluations list
        :type source_evaluation: List[Dict[str, Any]]
        :return: Parsed source evaluation
        :rtype: SourceEvaluation
        """
        return SourceEvaluation(
            commands_evaluations=[
                CommandEvaluation.from_dict(command_evaluation)
                for command_evaluation in source_evaluation["commands_evaluations"]
            ],
            source_execution_duration=source_evaluation["source_execution_duration"],
        )

    def __iter__(self) -> Iterator[CommandEvaluation]:
        """
        Iterate over commands evaluations.

        :return: commands evaluation iterator
        :rtype: Iterator[CommandEvaluation]
        """
        return iter(self.commands_evaluations)


@dataclass
class Evaluation:
    """Full evaluation class."""

    timestamp: datetime.datetime = field(
        default_factory=lambda: datetime.datetime.now().replace(microsecond=0)
    )
    sources_evaluations: Dict[Path, SourceEvaluation] = field(default_factory=dict)
    total_execution_duration: float = field(default=0)

    def __iter__(self) -> Iterator[Path]:
        """
        Iterate over evaluation.

        :return: self iterator
        :rtype: Iterator[Path]
        """
        return iter(self.sources_evaluations)

    def __getitem__(self, item: Path) -> SourceEvaluation:
        """
        Get source evaluation.

        :param item: Source name
        :type item: Path
        :return: Source's evaluation
        :rtype: SourceEvaluation
        """
        return self.sources_evaluations[item]

    def __setitem__(self, key: Path, value: SourceEvaluation) -> None:
        """
        Set source evaluation.

        :param key: Source name
        :type key: Path
        :param value: Source's evaluation
        :type value: SourceEvaluation
        """
        self.sources_evaluations[key] = value

    def keys(self) -> KeysView[Path]:
        """
        Get sources as generator.

        :return: All sources names generator
        :rtype: KeysView[str]
        """
        return self.sources_evaluations.keys()

    def values(self) -> ValuesView[SourceEvaluation]:
        """
        Get all source evaluations.

        :return: source evaluations list
        :rtype: ValuesView[SourceEvaluation]
        """
        return self.sources_evaluations.values()

    def items(self) -> ItemsView[Path, SourceEvaluation]:
        """
        Get sources evaluations.

        :return: All sources names-instance tuples generator
        :rtype: ItemsView[str, SourceEvaluation]
        """
        return self.sources_evaluations.items()

    def as_dict(self) -> Dict[str, Any]:
        """
        Return evaluation as json dictionary.

        :return: Self as dictionary
        :rtype: Dict[str, List[Dict[str, Any]]]
        """
        sources_evaluations = {str(key): value.as_dict() for key, value in self.items()}
        return dict(
            timestamp=self.timestamp.strftime(DATETIME_FORMAT),
            sources_evaluations=sources_evaluations,
            total_execution_duration=self.total_execution_duration,
        )

    def save_as_json(self, output: Path) -> None:
        """
        Save evaluation as json.

        :param output: Path to save self in
        :type output: Path
        """
        with open(output, mode="w", encoding=ENCODING) as output_file:
            json.dump(self.as_dict(), output_file, indent=2)

    @property
    def success(self) -> bool:
        """
        All sources evaluations are successful.

        :return: Success statue
        :rtype: bool
        """
        return all(
            sources_evaluation.success
            for sources_evaluation in self.sources_evaluations.values()
        )

    @property
    def commands_number(self):
        """
        Number of commands that were evaluated.

        :return: Counted commands
        :rtype: int
        """
        return sum(
            source_evaluation.commands_number
            for source_evaluation in self.sources_evaluations.values()
        )

    @property
    def successful_commands_number(self):
        """
        Number of successful commands that were evaluated.

        :return: Counted successful commands
        :rtype: int
        """
        return sum(
            source_evaluation.successful_commands_number
            for source_evaluation in self.sources_evaluations.values()
        )

    @property
    def failed_commands_number(self):
        """
        Number of failed commands that were evaluated.

        :return: Counted failed commands
        :rtype: int
        """
        return self.commands_number - self.successful_commands_number

    @classmethod
    def load_from_file(cls, input_path: Path) -> "Evaluation":
        """
        Load evaluation from json file.

        :param input_path: Path to load evaluation from.
        :type input_path: Path
        :return: Evaluation instance
        :rtype: Evaluation
        """
        with open(input_path, mode="r", encoding=ENCODING) as input_file:
            return Evaluation.from_dict(json.load(input_file))

    @property
    def commands_map(self) -> CommandsMap:
        """
        Get a map from input paths to commands.

        :return: Map of commands
        :rtype: CommandsMap
        """
        return CommandsMap(
            {
                input_path: [
                    command_evaluation.command
                    for command_evaluation in source_valuation.commands_evaluations
                ]
                for input_path, source_valuation in self.items()
            }
        )

    @property
    def failure_evaluation(self) -> "Evaluation":
        """
        Returns a new evaluation map with only failed commands.

        :return: Map of failed commands
        :rtype: Evaluation
        """
        failure_dict = {
            source: SourceEvaluation(
                [
                    command_evaluation
                    for command_evaluation in source_evaluation.commands_evaluations
                    if not command_evaluation.success
                ]
            )
            for source, source_evaluation in self.items()
        }
        failure_dict = {
            source: source_evaluation
            for source, source_evaluation in failure_dict.items()
            if len(source_evaluation) != 0
        }
        return Evaluation(
            timestamp=self.timestamp,
            sources_evaluations=failure_dict,
            total_execution_duration=self.total_execution_duration,
        )

    @classmethod
    def from_dict(cls, evaluation: Dict[str, Any]) -> "Evaluation":
        """
        Read evaluation from json dictionary.

        :param evaluation: Json evaluation
        :type evaluation: List[Dict[str, Any]]]
        :return: Parsed evaluation
        :rtype: Evaluation
        """
        return Evaluation(
            timestamp=datetime.datetime.strptime(
                evaluation["timestamp"], DATETIME_FORMAT
            ),
            sources_evaluations={
                Path(input_path): SourceEvaluation.from_dict(source_evaluation)
                for input_path, source_evaluation in evaluation[
                    "sources_evaluations"
                ].items()
            },
            total_execution_duration=evaluation["total_execution_duration"],
        )



"""Place for saving all available contexts."""
import itertools
from collections import OrderedDict
from typing import Any, Dict, Iterator, List
from typing import OrderedDict as OrderedDictType

from statue.constants import ALIASES, ALLOWED_BY_DEFAULT, HELP, PARENT
from statue.context import Context
from statue.exceptions import (
    InconsistentConfiguration,
    MissingHelpString,
    UnknownContext,
)


class ContextsRepository:
    """Repository class for saving and accessing contexts."""

    def __init__(self, *contexts: Context):
        """
        Constructor.

        :param contexts: Initial contexts to be saved
        :type contexts: Context
        """
        self.contexts_list = list(contexts)

    def __len__(self) -> int:
        """
        Number of available contexts.

        :return: Number of contexts
        :rtype: int
        """
        return len(self.contexts_list)

    def __iter__(self) -> Iterator[Context]:
        """
        Iterate over all available contexts.

        :return: Contexts iterator
        :rtype: Iterator[Context]
        """
        return iter(self.contexts_list)

    def __getitem__(self, item: str) -> Context:
        """
        Get context by name or alias.

        :param item: Context name or alias to be retrieved
        :type item: str
        :return: Context with given name
        :rtype: Context
        :raises UnknownContext: Raised when context is not found
        """
        for context in self.contexts_list:
            if context.is_matching(item):
                return context
        raise UnknownContext(context_name=item)

    def __contains__(self, item: str) -> bool:
        """
        Does context available in repository.

        :param item: Context name to be searched
        :type item: str
        :return: does context exist in repository
        :rtype: bool
        """
        return item in self.occupied_names

    @property
    def occupied_names(self) -> List[str]:
        """List of all occupied names in contexts repository."""
        occupied_names = list(
            itertools.chain.from_iterable([context.all_names for context in self])
        )
        occupied_names.sort()
        return occupied_names

    def add_contexts(self, *contexts: Context):
        """
        Add contexts to repository.

        :param contexts: Contexts to be added to the repository
        :type contexts: Context
        :raises InconsistentConfiguration: Raised when trying to add contexts with the
            same name or an existing name.
        """
        message = "context name or alias has been defined twice"
        for i, context in enumerate(contexts):
            existing_aliases = [alias for alias in context.all_names if alias in self]
            if len(existing_aliases) != 0:
                raise InconsistentConfiguration(message, location=[existing_aliases[0]])
            for j in range(i):
                other_context = contexts[j]
                overlapping_aliases = [
                    context_alias
                    for context_alias in context.all_names
                    if context_alias in other_context.all_names
                ]
                if len(overlapping_aliases) != 0:
                    raise InconsistentConfiguration(
                        message, location=[overlapping_aliases[0]]
                    )
        self.contexts_list.extend(contexts)

    def remove_context(self, context: Context):
        """
        Remove a context from the repository.

        :param context: Context to be removed from the repository
        :type context: Context
        """
        self.contexts_list.remove(context)

    def reset(self):
        """Clear repository from all contexts."""
        self.contexts_list.clear()

    def as_dict(self) -> OrderedDictType[str, Any]:
        """
        Encode contexts repository as a dictionary.

        This is used in order to serialize the contexts repository in
        a configuration file.

        :return: Serialized representation dictionary
        :rtype: OrderedDict[str, Any]
        """
        contexts_list = list(self)
        contexts_list.sort(key=lambda context: context.name)
        return OrderedDict(
            [(context.name, context.as_dict()) for context in contexts_list]
        )

    @classmethod
    def from_dict(cls, config: Dict[str, Any]) -> "ContextsRepository":
        """
        Create contexts repository from given configuration.

        :param config: Configuration to update repository from
        :type config: Dict[str, Any]
        :return: Contexts repository object.
        :rtype: ContextsRepository
        :raises InconsistentConfiguration: Raised when inconsistency is found in
            configuration.
        """
        contexts_repository = ContextsRepository()
        context_names = set(config.keys())
        while len(context_names) != 0:
            available_contexts = [
                context_name
                for context_name in context_names
                if cls._can_be_built(
                    context_config=config[context_name],
                    contexts_repository=contexts_repository,
                )
            ]
            if len(available_contexts) == 0:
                raise InconsistentConfiguration(
                    "The following contexts cannot be built because they are missing "
                    "or they cause circular parenting: "
                    f"{', '.join(context_names)}"
                )
            for context_name in available_contexts:
                context_names.remove(context_name)
                cls._add_context_from_config(
                    context_name=context_name,
                    context_config=config[context_name],
                    contexts_repository=contexts_repository,
                )
        return contexts_repository

    @classmethod
    def _add_context_from_config(
        cls,
        context_name: str,
        context_config: Dict[str, Any],
        contexts_repository: "ContextsRepository",
    ):
        """
        Add context from given configuration to contexts repository.

        If context is not available in configuration, try get it
        from the repository (if it exists).

        :param context_name: Context to be built
        :type context_name: str
        :param context_config: Configuration to build context from
        :type context_config: Dict[str, Any]
        :param contexts_repository: Contexts repository to add new context to
        :type contexts_repository: ContextsRepository
        :raises MissingHelpString: Raised when help string is missing
        """
        parent = (
            contexts_repository[context_config[PARENT]]
            if PARENT in context_config
            else None
        )
        aliases = context_config.get(ALIASES, [])
        if HELP not in context_config:
            raise MissingHelpString(location=[context_name])
        contexts_repository.add_contexts(
            Context(
                name=context_name,
                help=context_config[HELP],
                aliases=aliases,
                parent=parent,
                allowed_by_default=context_config.get(ALLOWED_BY_DEFAULT, False),
            )
        )

    @classmethod
    def _can_be_built(
        cls, context_config: Dict[str, Any], contexts_repository: "ContextsRepository"
    ) -> bool:
        if PARENT not in context_config:
            return True
        if context_config[PARENT] in contexts_repository:
            return True
        return False



"""Commands CLI."""
import click

from statue.cli.cli import pass_configuration, statue_cli
from statue.cli.common_flags import silent_option, verbose_option, verbosity_option
from statue.cli.styled_strings import bullet_style, name_style
from statue.config.configuration import Configuration
from statue.exceptions import UnknownCommand


@statue_cli.group("commands")
def commands_cli() -> None:
    """Commands related actions such as list, install, show, etc."""


@commands_cli.command("list")
@pass_configuration
def list_commands_cli(configuration: Configuration) -> None:
    """List matching commands to contexts, allow list and deny list."""
    for command_builder in configuration.commands_repository:
        name = name_style(command_builder.name)
        if command_builder.version is not None:
            name += f" (version: {command_builder.version})"
        click.echo(f"{name} - {command_builder.help}")


@commands_cli.command("install")
@silent_option
@verbose_option
@verbosity_option
@pass_configuration
def install_commands_cli(
    configuration: Configuration,
    verbosity: str,
) -> None:
    """Install missing commands."""
    for command_builder in configuration.commands_repository:
        if not command_builder.installed_correctly():
            command_builder.install(verbosity=verbosity)


@commands_cli.command("show")
@click.argument("command_name", type=str)
@click.pass_context
@pass_configuration
def show_command_cli(
    configuration: Configuration,
    ctx: click.Context,
    command_name: str,
) -> None:
    """Show information about specific command."""
    command_builder = None
    try:
        command_builder = configuration.commands_repository[command_name]
    except UnknownCommand as error:
        click.echo(str(error))
        ctx.exit(1)
    click.echo(f"{bullet_style('Name')} - {name_style(command_builder.name)}")
    if command_builder.version is not None:
        click.echo(f"{bullet_style('Version')} - {command_builder.version}")
    click.echo(f"{bullet_style('Description')} - {command_builder.help}")
    if len(command_builder.default_args) != 0:
        click.echo(
            f"{bullet_style('Default arguments')} - "
            f"{' '.join(command_builder.default_args)}"
        )
    if len(command_builder.required_contexts) != 0:
        required_contexts = [
            name_style(context.name) for context in command_builder.required_contexts
        ]
        required_contexts.sort()
        click.echo(
            f"{bullet_style('Required contexts')} - " f"{', '.join(required_contexts)}"
        )
    if len(command_builder.allowed_contexts) != 0:
        allowed_contexts = [
            name_style(context.name) for context in command_builder.allowed_contexts
        ]
        allowed_contexts.sort()
        click.echo(
            f"{bullet_style('Allowed contexts')} - " f"{', '.join(allowed_contexts)}"
        )
    if len(command_builder.denied_contexts) != 0:
        denied_contexts = [
            name_style(context.name) for context in command_builder.denied_contexts
        ]
        denied_contexts.sort()
        click.echo(
            f"{bullet_style('Denied contexts')} - " f"{', '.join(denied_contexts)}"
        )
    if len(command_builder.specified_contexts) == 0:
        return
    click.echo(f"{bullet_style('Specified contexts')}:")
    for (
        context,
        context_specification,
    ) in command_builder.contexts_specifications.items():
        click.echo(f"\t{name_style(context.name)}")
        if context_specification.args is not None:
            click.echo(
                f"\t\t{bullet_style('Override arguments')}: "
                f"{' '.join(context_specification.args)}"
            )
        if context_specification.add_args is not None:
            click.echo(
                f"\t\t{bullet_style('Added arguments')}: "
                f"{' '.join(context_specification.add_args)}"
            )
        if context_specification.clear_args:
            click.echo(f"\t\t{bullet_style('Clears arguments')}")



# pylint: disable=too-many-locals
"""Run CLI."""
from pathlib import Path
from typing import List, Optional, Sequence

import click

from statue.cli.cli import pass_configuration, statue_cli
from statue.cli.cli_util import list_or_none
from statue.cli.common_flags import (
    allow_option,
    contexts_option,
    deny_option,
    silent_option,
    verbose_option,
    verbosity_option,
)
from statue.cli.string_util import (
    boxed_string,
    evaluation_string,
    evaluation_summary_string,
)
from statue.cli.styled_strings import failure_style
from statue.commands_map_builder import CommandsMapBuilder
from statue.config.configuration import Configuration
from statue.exceptions import CommandsMapBuilderError, UnknownContext
from statue.runner import RunnerMode, build_runner
from statue.verbosity import is_silent, is_verbose


@statue_cli.command("run", short_help="Run static code analysis commands on sources.")
@click.argument("sources", type=click.Path(exists=True, path_type=Path), nargs=-1)
@contexts_option
@allow_option
@deny_option
@click.option(
    "-i", "--install", is_flag=True, help="Install commands before running if missing"
)
@click.option(
    "-p",
    "--previous",
    type=int,
    help=(
        "Run commands of the nth recent evaluation. "
        'combine this flag with "-f" in order to run only failed commands from that '
        "evaluation"
    ),
)
@click.option(
    "-r",
    "--recent",
    "previous",
    flag_value=1,
    type=int,
    help='Run commands of the most recent evaluation. Same as "--previous 1".',
)
@click.option(
    "-f", "--failed", is_flag=True, help="Run commands from the most recent failed run"
)
@click.option(
    "-fo",
    "--failed-only",
    is_flag=True,
    help=(
        'Same as "--failed", but will only run the commands '
        "and not those who ended successfully"
    ),
)
@click.option(
    "--cache/--no-cache", default=True, help="Save evaluation to cache or not"
)
@silent_option
@verbose_option
@verbosity_option
@click.option(
    "--mode",
    type=click.Choice([mode.name.lower() for mode in RunnerMode], case_sensitive=False),
    callback=lambda ctx, param, value: (None if value is None else value.upper()),
    help="Should run asynchronously or not.",
)
@click.option(
    "-o",
    "--output",
    type=click.Path(dir_okay=False, path_type=Path),
    help="Output path to save evaluation result",
)
@click.pass_context
@pass_configuration
def run_cli(  # pylint: disable=too-many-arguments
    configuration: Configuration,
    ctx: click.Context,
    sources: Sequence[Path],
    context: Sequence[str],
    allow: Optional[List[str]],
    deny: Optional[List[str]],
    previous: Optional[int],
    failed: bool,
    failed_only: bool,
    install: bool,
    cache: bool,
    verbosity: str,
    mode: Optional[str],
    output: Optional[Path],
) -> None:
    """
    Run static code analysis commands on sources.

    Source files to run Statue on can be presented as positional arguments.
    When no source files are presented, will use configuration file to determine on
    which files to run
    """
    commands_map = None
    try:
        commands_map = CommandsMapBuilder(
            configuration=configuration,
            specified_sources=list_or_none(sources),
            allowed_commands=list_or_none(allow),
            denied_commands=list_or_none(deny),
            contexts=[
                configuration.contexts_repository[context_name]
                for context_name in context
            ],
            previous=previous,
            failed=failed,
            failed_only=failed_only,
        ).build()
    except (UnknownContext, CommandsMapBuilderError) as error:
        click.echo(failure_style(str(error)))
        ctx.exit(1)
    if len(commands_map) == 0:
        click.echo("No commands to run.")
        ctx.exit(0)
    missing_commands = [
        command_builder
        for command_builder in configuration.commands_repository
        if command_builder.name in commands_map.command_names
        and not command_builder.installed_correctly()
    ]
    __handle_missing_commands(
        ctx=ctx,
        missing_commands=missing_commands,
        install=install,
        verbosity=verbosity,
    )
    mode = mode if mode is not None else configuration.default_mode.name
    if is_verbose(verbosity):
        click.echo(f"Running evaluation in {mode.lower()} mode")
    runner = build_runner(mode)
    evaluation = runner.evaluate(commands_map)
    if not is_silent(verbosity):
        click.echo(boxed_string("Evaluation"))
        click.echo(evaluation_string(evaluation, verbosity=verbosity))
    if cache and configuration.cache.enabled:
        configuration.cache.save_evaluation(evaluation)
    if output is not None:
        evaluation.save_as_json(output)
    click.echo()
    if not is_silent(verbosity):
        click.echo(boxed_string("Summary"))
        click.echo()
    click.echo(evaluation_summary_string(evaluation))
    exit_code = 0 if evaluation.success else 1
    ctx.exit(exit_code)


def __handle_missing_commands(ctx, missing_commands, install, verbosity):
    if len(missing_commands) == 0:
        return
    if install:
        for command in missing_commands:
            command.update_to_version(verbosity=verbosity)
    else:
        missing_commands_names = [command.name for command in missing_commands]
        click.echo(
            failure_style(
                "The following commands are not installed correctly: "
                f"{', '.join(missing_commands_names)}"
            )
        )
        click.echo(
            "Consider using the '-i' flag in order to install missing "
            "commands before running"
        )
        ctx.exit(1)



"""Common flags used by multiple CLIs."""
from pathlib import Path

import click

from statue.verbosity import DEFAULT_VERBOSITY, SILENT, VERBOSE, VERBOSITIES

config_path_option = click.option(
    "--config",
    envvar="STATUE_CONFIG",
    type=click.Path(exists=True, dir_okay=False, path_type=Path),
    help="Statue configuration file.",
)
contexts_option = click.option(
    "-c",
    "--context",
    type=str,
    default=None,
    multiple=True,
    help="Context in which to evaluate the commands.",
)
allow_option = click.option(
    "-a", "--allow", type=str, default=None, multiple=True, help="Allowed command."
)
deny_option = click.option(
    "-d", "--deny", type=str, default=None, multiple=True, help="Denied command."
)
verbosity_option = click.option(
    "--verbosity",
    type=click.Choice(VERBOSITIES, case_sensitive=False),
    default=DEFAULT_VERBOSITY,
    show_default=True,
)

silent_option = click.option(
    "--silent", "verbosity", flag_value=SILENT, help=f'Set verbosity to "{SILENT}".'
)

verbose_option = click.option(
    "--verbose", "verbosity", flag_value=VERBOSE, help=f'Set verbosity to "{VERBOSE}".'
)



"""General configuration CLI."""
import click

from statue.cli.common_flags import config_path_option
from statue.cli.config.config_cli import config_cli
from statue.config.configuration import Configuration
from statue.runner import RunnerMode


@config_cli.command("set-mode")
@click.argument(
    "mode",
    type=click.Choice([mode.name.lower() for mode in RunnerMode], case_sensitive=False),
    callback=lambda ctx, param, value: (None if value is None else value.upper()),
)
@config_path_option
def set_mode_cli(mode, config):
    """Choose which runner mode will be used by default."""
    if config is None:
        config = Configuration.configuration_path()
    configuration = Configuration.from_file(config)
    configuration.default_mode = RunnerMode[mode]
    configuration.to_toml(config)
    click.echo("Mode was successfully set!")


@config_cli.command("set-history-size")
@click.argument("size", type=int, nargs=1)
@config_path_option
def set_history_size_cli(size, config):
    """Choose which runner mode will be used by default."""
    if config is None:
        config = Configuration.configuration_path()
    configuration = Configuration.from_file(config)
    configuration.cache.history_size = size
    configuration.to_toml(config)
    click.echo("History size was successfully set!")


@config_cli.command("enable-cache")
@config_path_option
def enable_cache_cli(config):
    """Enabled caching in configuration."""
    if config is None:
        config = Configuration.configuration_path()
    configuration = Configuration.from_file(config)
    configuration.cache.enabled = True
    configuration.to_toml(config)
    click.echo("Caching is enabled!")


@config_cli.command("disable-cache")
@config_path_option
def disable_cache_cli(config):
    """Disabled caching in configuration."""
    if config is None:
        config = Configuration.configuration_path()
    configuration = Configuration.from_file(config)
    configuration.cache.enabled = False
    configuration.to_toml(config)
    click.echo("Caching is disabled!")



"""Module for cache related methods."""
from collections import deque
from pathlib import Path
from typing import Deque, List, Optional, Sequence, Set

from statue.evaluation import Evaluation
from statue.exceptions import CacheError


class Cache:
    """Cache files repository."""

    def __init__(
        self,
        size: int,
        cache_root_directory: Optional[Path] = None,
        enabled: bool = True,
    ):
        """
        Initialize cache.

        :param size: Number of evaluations to save
        :type size: int
        :param cache_root_directory: Optional root directory for caching
        :type cache_root_directory: Optional[Path]
        :param enabled: Whether caching is enabled or not. True by default.
        :type enabled: bool
        """
        self._all_evaluations: Deque[Evaluation] = deque()
        self.cache_root_directory = cache_root_directory
        self.history_size = size
        self.enabled = enabled

    @property
    def cache_root_directory(self) -> Optional[Path]:
        """Root directory for caching."""
        return self._cache_root_directory

    @cache_root_directory.setter
    def cache_root_directory(self, cache_dir: Optional[Path]):
        """
        Set root directory for caching.

        :param cache_dir: Root directory for caching
        :type cache_dir: Optional[Path]
        """
        self._cache_root_directory = cache_dir
        if cache_dir is None:
            return
        self.__ensure_dir_exists(cache_dir)
        self.load_evaluations()

    @property
    def evaluations_dir(self) -> Optional[Path]:
        """
        Directory of cache files. Created if missing.

        :return: Location path of the previous evaluations cache directory
        :rtype: Path
        """
        if self.cache_root_directory is None:
            return None
        return self.__ensure_dir_exists(self.cache_root_directory / "evaluations")

    @property
    def all_evaluation_paths(self) -> Set[Path]:
        """
        Get all evaluation paths, ordered from recent to last.

        :return: Set of all previous evaluations paths
        :rtype: Set[Path]
        """
        if self.evaluations_dir is None:
            return set()
        return set(self.evaluations_dir.iterdir())

    @property
    def all_evaluations(self) -> List[Evaluation]:
        """All cached evaluations."""
        return list(self._all_evaluations)

    @all_evaluations.setter
    def all_evaluations(self, evaluations: Sequence[Evaluation]):
        self.all_evaluations.clear()
        evaluations = list(evaluations)
        evaluations.sort(key=lambda evaluation: evaluation.timestamp)
        self._all_evaluations.extendleft(evaluations)

    @property
    def number_of_evaluations(self) -> int:
        """Get number of cached evaluations."""
        return len(self.all_evaluations)

    @property
    def recent_failed_evaluation(self) -> Evaluation:
        """Get the most recent failed evaluation."""
        for evaluation in self.all_evaluations:
            if not evaluation.success:
                return evaluation
        raise CacheError("Could not find failed evaluation")

    def get_evaluation(self, n: int) -> Evaluation:
        """
        Get the nth most recent evaluation.

        :param n: Evaluation index
        :type n: int
        :return: The nth evaluation
        :rtype: Evaluation
        :raises CacheError: raised when receiving an invalid index for evaluation
        """
        if n < 0 or n >= self.number_of_evaluations:
            raise CacheError(
                "Could not get the desired evaluation due to invalid index"
            )
        return self.all_evaluations[n]

    def save_evaluation(self, evaluation: Evaluation):
        """
        Save evaluation to cache.

        Deletes old evaluations after saving according to history size.

        :param evaluation: Evaluation instance to be saved
        :type evaluation: Evaluation
        """
        self._all_evaluations.appendleft(evaluation)
        evaluation.save_as_json(self.__get_evaluation_path(evaluation))
        while len(self.all_evaluations) > self.history_size:
            self.__remove_oldest_evaluation()

    def clear(self, limit: Optional[int] = None):
        """
        Remove evaluations from cache.

        :param limit: Optional. limit the number of evaluations to be deleted
        :type limit: Optional[int]
        """
        number_of_evaluations_to_be_deleted = (
            limit if limit is not None else self.number_of_evaluations
        )
        for _ in range(number_of_evaluations_to_be_deleted):
            self.__remove_oldest_evaluation()

    def load_evaluations(self):
        """Load all evaluations from evaluations directory."""
        self.all_evaluations = [
            Evaluation.load_from_file(evaluation_path)
            for evaluation_path in self.all_evaluation_paths
        ]

    def __remove_oldest_evaluation(self):
        evaluation = self._all_evaluations.pop()
        self.__get_evaluation_path(evaluation).unlink()

    def __get_evaluation_path(self, evaluation: Evaluation) -> Path:
        if self.evaluations_dir is None:
            raise CacheError("Cache directory was not specified")
        seconds_since_epoch = int(evaluation.timestamp.timestamp())
        return self.evaluations_dir / f"evaluation-{seconds_since_epoch}.json"

    @classmethod
    def __ensure_dir_exists(cls, dir_path: Path) -> Path:
        dir_path.mkdir(parents=True, exist_ok=True)
        return dir_path



"""Place for saving all available sources with default commands filter."""
from collections import OrderedDict
from pathlib import Path
from typing import Any, Dict, List, Optional
from typing import OrderedDict as OrderedDictType

from statue.commands_filter import CommandsFilter
from statue.config.contexts_repository import ContextsRepository
from statue.constants import ALLOW_LIST, CONTEXTS, DENY_LIST
from statue.io_util import is_equal_or_child_of


class SourcesRepository:
    """
    Repository class for saving and accessing sources.

    This is done in order to get specific commands filter from each source.
    """

    def __init__(
        self, sources_filters_map: Optional[Dict[Path, CommandsFilter]] = None
    ):
        """
        Initialize repository.

        :param sources_filters_map: Map from a path to its commands filter
        :type sources_filters_map: Optional[Dict[Path, CommandsFilter]]
        """
        self.sources_filters_map: Dict[Path, CommandsFilter] = (
            {} if sources_filters_map is None else sources_filters_map
        )

    def __len__(self) -> int:
        """
        Number of available sources commands filters.

        :return: Number of sources commands filters
        :rtype: int
        """
        return len(self.sources_filters_map)

    def __setitem__(self, key: Path, value: CommandsFilter):
        """
        Set a commands filter for a given source.

        :param key: Source to save commands filter for
        :type key: Path
        :param value: Command filter for given source
        :type value: CommandsFilter
        """
        self.sources_filters_map[key] = value

    def __getitem__(self, item: Path) -> CommandsFilter:
        """
        Get a commands filter for a given source.

        If commands filter is defined for given source, returns default
        filter.

        :param item: Source to get commands filter for
        :type item: Path
        :return: Command filter for given source
        :rtype: CommandsFilter
        """
        for specified_source, commands_filter in self.sources_filters_map.items():
            if is_equal_or_child_of(item, specified_source):
                return commands_filter
        return CommandsFilter()

    @property
    def sources_list(self) -> List[Path]:
        """Get list of all available sources."""
        return list(self.sources_filters_map.keys())

    def track_sources(self, *sources: Path):
        """
        Track sources with default commands filter.

        :param sources: Sources to track
        :type sources: Path
        """
        for source in sources:
            self[source] = CommandsFilter()

    def remove_source(self, source: Path):
        """
        Remove source from repository.

        :param source: Source to be removed
        :type source Path
        """
        del self.sources_filters_map[source]

    def reset(self):
        """Reset sources repository."""
        self.sources_filters_map.clear()

    def as_dict(self) -> OrderedDictType[str, Any]:
        """
        Encode sources repository as a dictionary.

        This is used in order to serialize the sources repository in
        a configuration file.

        :return: Serialized representation dictionary
        :rtype: Dict[str, Any]
        """
        sources_list = self.sources_list
        sources_list.sort(key=lambda source: source.as_posix())
        return OrderedDict(
            [(source.as_posix(), self[source].as_dict()) for source in sources_list]
        )

    @classmethod
    def from_dict(cls, config: Dict[str, Any], contexts_repository: ContextsRepository):
        """
        Create sources repository from configuration mapping.

        This is done using an existing contexts repository to link contexts
        from config to existing context objects


        :param config: Configuration to create repository from
        :type config: Dict[str, Any]
        :param contexts_repository: Contexts repository to get context objects
            from
        :type contexts_repository: ContextsRepository
        :return: Built sources repository
        :rtype: SourcesRepository
        """
        sources_filters_map = {}
        for source, commands_filter_config in config.items():
            contexts = [
                contexts_repository[context_name]
                for context_name in commands_filter_config.get(CONTEXTS, [])
            ]
            sources_filters_map[Path(source)] = CommandsFilter(
                contexts=contexts,
                allowed_commands=commands_filter_config.get(ALLOW_LIST, None),
                denied_commands=commands_filter_config.get(DENY_LIST, None),
            )
        return SourcesRepository(sources_filters_map)



"""Commands map allow us to know which commands to run on each source."""
import itertools
from typing import Set


class CommandsMap(dict):
    """A mapping from source path to commands to run on it."""

    @property
    def total_commands_count(self) -> int:
        """How many commands total in the commands map."""
        return sum(len(value) for value in self.values())

    @property
    def command_names(self) -> Set[str]:
        """All command names in map."""
        return set(
            itertools.chain.from_iterable(
                [command.name for command in sources_commands]
                for sources_commands in self.values()
            )
        )



"""Contexts CLI."""

import click

from statue.cli.cli import pass_configuration, statue_cli
from statue.cli.styled_strings import bullet_style, name_style
from statue.config.configuration import Configuration
from statue.exceptions import UnknownContext


@statue_cli.group("contexts")
def context_cli() -> None:
    """Contexts related actions such as list, show, etc."""


@context_cli.command("list")
@click.pass_context
@pass_configuration
def contexts_list_cli(configuration: Configuration, ctx: click.Context) -> None:
    """Print all available contexts."""
    if len(configuration.contexts_repository) == 0:
        click.echo("No contexts were found.")
        ctx.exit(1)
    for context in configuration.contexts_repository:
        click.echo(f"{name_style(context.name)} - {context.help}")


@context_cli.command("show")
@click.argument("context_name", type=str)
@click.pass_context
@pass_configuration
def show_contexts_cli(
    configuration: Configuration, ctx: click.Context, context_name: str
) -> None:
    """Print all available contexts."""
    try:
        context = configuration.contexts_repository[context_name]
        click.echo(f"{bullet_style('Name')} - {name_style(context.name)}")
        click.echo(f"{bullet_style('Description')} - {context.help}")
        if len(context.aliases) != 0:
            click.echo(f"{bullet_style('Aliases')} - {', '.join(context.aliases)}")
        if context.parent is not None:
            click.echo(f"{bullet_style('Parent')} - {name_style(context.parent.name)}")
        required_by = [
            name_style(command_builder.name)
            for command_builder in configuration.commands_repository
            if any(
                required_context == context
                for required_context in command_builder.required_contexts
            )
        ]
        if len(required_by) != 0:
            click.echo(f"{bullet_style('Required by')} - {', '.join(required_by)}")
        allowed_for = [
            name_style(command_builder.name)
            for command_builder in configuration.commands_repository
            if any(
                allowed_context == context
                for allowed_context in command_builder.allowed_contexts
            )
        ]
        if len(allowed_for) != 0:
            click.echo(f"{bullet_style('Allowed for')} - {', '.join(allowed_for)}")
        specified_for = [
            name_style(command_builder.name)
            for command_builder in configuration.commands_repository
            if any(
                specified_context == context
                for specified_context in command_builder.specified_contexts
            )
        ]
        if len(specified_for) != 0:
            click.echo(f"{bullet_style('Specified for')} - {', '.join(specified_for)}")
    except UnknownContext:
        click.echo(f'Could not find the context "{context_name}".')
        ctx.exit(1)



"""Singleton for storing sources locks for async commands execution."""
import asyncio
from pathlib import Path
from typing import Dict


class SourcesLocksRepository:  # pylint: disable=too-few-public-methods
    """
    Singleton for storing sources locks.

    This is used for async commands executions.
    """

    total_lock: asyncio.Lock = asyncio.Lock()
    locks_dict: Dict[Path, asyncio.Lock] = {}

    @classmethod
    async def get_lock(cls, source: Path) -> asyncio.Lock:
        """
        Get lock of a specific source.

        First check if the source has an existing lock. For this check
        we must lock the entire repository using the total_lock.
        If lock is not existing, creating it.

        After that, returning the lock.

        :param source: The source to get its lock
        :type source: Path
        :return: The source's lock
        :rtype: asyncio.Lock
        """
        await cls.total_lock.acquire()
        if source not in cls.locks_dict:
            cls.locks_dict[source] = asyncio.Lock()
        cls.total_lock.release()
        return cls.locks_dict[source]



"""Configuration related CLIs."""
from statue.cli.config.config_cli import config_cli
from statue.cli.config.config_commands import (
    add_command_cli,
    edit_command_cli,
    fixate_commands_versions_cli,
    remove_command_cli,
)
from statue.cli.config.config_contexts import (
    add_context_cli,
    edit_context_cli,
    remove_context_cli,
)
from statue.cli.config.config_general import set_history_size_cli, set_mode_cli
from statue.cli.config.config_init import init_config_cli
from statue.cli.config.config_show import show_config_cli
from statue.cli.config.config_sources import (
    add_source_to_configuration_cli,
    edit_source_in_configuration_cli,
    remove_source_from_configuration_cli,
)
from statue.cli.show_tree import show_tree_cli

__all__ = [
    "config_cli",
    "set_mode_cli",
    "set_history_size_cli",
    "add_context_cli",
    "edit_context_cli",
    "remove_context_cli",
    "add_command_cli",
    "edit_command_cli",
    "remove_command_cli",
    "fixate_commands_versions_cli",
    "add_source_to_configuration_cli",
    "edit_source_in_configuration_cli",
    "remove_source_from_configuration_cli",
    "init_config_cli",
    "show_config_cli",
    "show_tree_cli",
]



"""History CLI."""
import sys
from datetime import datetime
from typing import Optional, Union

import click

from statue.cli.cli import pass_configuration, statue_cli
from statue.cli.common_flags import verbose_option
from statue.cli.styled_strings import (
    bullet_style,
    failure_style,
    name_style,
    source_style,
    success_style,
)
from statue.command import CommandEvaluation
from statue.config.configuration import Configuration
from statue.constants import DATETIME_FORMAT
from statue.evaluation import Evaluation
from statue.exceptions import CacheError
from statue.verbosity import is_verbose


def evaluation_status(evaluation: Union[Evaluation, CommandEvaluation]) -> str:
    """
    Get styled evaluation string.

    :param evaluation: The evaluation to get the status of
    :type evaluation: Evaluation or CommandEvaluation
    :return: styles success/failure string
    :rtype: str
    """
    if evaluation.success:
        return success_style("Success")
    return failure_style("Failure")


def evaluation_success_ratio(evaluation: Evaluation) -> str:
    """
    Get evaluation ratio string.

    :param evaluation: The evaluation to get the status of
    :type evaluation: Evaluation
    :return: success ratio string
    :rtype: str
    """
    return f"{evaluation.successful_commands_number}/{evaluation.commands_number}"


def total_evaluation_string(evaluation: Evaluation) -> str:
    """
    Create a string representing an evaluation.

    :param evaluation: The actual evaluation instance.
    :type evaluation: Evaluation
    :return: Pretty string describing the evaluation
    :rtype: str
    """
    return (
        f"{bullet_style(datetime.strftime(evaluation.timestamp, DATETIME_FORMAT))} -"
        f" {evaluation_status(evaluation)} "
        f"({evaluation_success_ratio(evaluation)} successful, "
        f"{evaluation.total_execution_duration:.2f} seconds)"
    )


@statue_cli.group("history")
def history_cli() -> None:
    """History related actions such as list, show, etc."""


@history_cli.command("list")
@click.option("--head", type=int, help="Show only the nth recent evaluations")
@pass_configuration
def list_evaluations_cli(configuration: Configuration, head: int):
    """List all recent evaluations."""
    evaluations = configuration.cache.all_evaluations
    if len(evaluations) == 0:
        click.echo("No previous evaluations.")
        return
    if head is not None:
        evaluations = evaluations[:head]
    for i, evaluation in enumerate(evaluations, start=1):
        click.echo(f"{i}) {total_evaluation_string(evaluation)}")


@history_cli.command("show")
@click.option(
    "-n", "number", type=int, default=1, help="Show nth recent evaluation. 1 by default"
)
@pass_configuration
@verbose_option
def show_evaluation_cli(
    configuration: Configuration,
    number: int,
    verbosity: str,
):
    """Show past evaluation."""
    try:
        evaluation = configuration.cache.get_evaluation(number - 1)
    except CacheError:
        click.echo(
            failure_style(f"Could not find evaluation with given index {number}")
        )
        sys.exit(1)
    click.echo(total_evaluation_string(evaluation))
    for source, source_evaluation in evaluation.items():
        click.echo(
            f"{source_style(str(source))} ("
            f"{source_evaluation.source_execution_duration:.2f} seconds):"
        )
        for command_evaluation in source_evaluation.commands_evaluations:
            click.echo(
                f"\t{name_style(command_evaluation.command.name)} - "
                f"{evaluation_status(command_evaluation)} "
                f"({command_evaluation.execution_duration:.2f} seconds)"
            )
            if is_verbose(verbosity):
                click.echo(
                    f"\t\tArguments: {' '.join(command_evaluation.command.args)}"
                )


@history_cli.command("clear")
@click.option("-f", "--force", is_flag=True, help="Force deletion and avoid prompt.")
@click.option(
    "-l",
    "--limit",
    type=int,
    help="Limit the number of deleted records. Deletes earliest evaluations.",
)
@pass_configuration
def clear_history_cli(configuration: Configuration, force: bool, limit: Optional[int]):
    """Clear records of previous statue runs."""
    number_of_evaluations = configuration.cache.number_of_evaluations
    if number_of_evaluations == 0:
        click.echo("No previous evaluations.")
        return
    number_of_files_to_be_deleted = (
        limit if limit is not None else number_of_evaluations
    )
    if not force:
        confirmation = click.confirm(
            f"{number_of_files_to_be_deleted} evaluation files are "
            "about to be deleted. Are you sure you want to delete them?",
            default=False,
        )
        if not confirmation:
            click.echo("Aborted without clearing history.")
            return
    configuration.cache.clear(limit=limit)
    click.echo(
        f"{number_of_files_to_be_deleted} evaluation files "
        "have been deleted successfully."
    )



"""Designated class for building commands map."""
from dataclasses import dataclass, field
from pathlib import Path
from typing import List, Optional

from statue.commands_filter import CommandsFilter
from statue.commands_map import CommandsMap
from statue.config.configuration import Configuration
from statue.context import Context
from statue.evaluation import Evaluation
from statue.exceptions import CacheError, CommandsMapBuilderError


@dataclass
class CommandsMapBuilder:  # pylint: disable=too-many-instance-attributes
    """Commands map builder class."""

    configuration: Configuration
    specified_sources: Optional[List[Path]] = None
    allowed_commands: Optional[List[str]] = None
    denied_commands: Optional[List[str]] = None
    contexts: List[Context] = field(default_factory=list)
    previous: Optional[int] = None
    failed: bool = False
    failed_only: bool = False

    @property
    def default_filter(self):
        """Get default filters for commands map."""
        return CommandsFilter(
            contexts=self.contexts,
            allowed_commands=self.allowed_commands,
            denied_commands=self.denied_commands,
        )

    @property
    def build_from_cache(self) -> bool:
        """Should build commands map using cached evaluations."""
        return self.previous is not None or self.failed or self.failed_only

    def build(self) -> CommandsMap:
        """
        Build commands map using the builders settings.

        :return: Commands map from instruction
        :rtype: CommandsMap
        :raises CommandsMapBuilderError: Raised when no sources where specified
        """
        if self.failed and self.previous is not None:
            raise CommandsMapBuilderError(
                '"failed" and "previous" cannot both be set when building commands map'
            )
        if not self.build_from_cache:
            sources = self.get_sources()
            if len(sources) == 0:
                raise CommandsMapBuilderError(
                    "No source was specified and no Sources section in configuration."
                )
            return self.configuration.build_commands_map(
                sources=sources, commands_filter=self.default_filter
            )
        evaluation = self.get_evaluation_from_cache()
        if self.failed_only:
            return evaluation.failure_evaluation.commands_map
        return evaluation.commands_map

    def get_evaluation_from_cache(self) -> Evaluation:
        """
        Get evaluation from cache based on flags.

        :return: Desired evaluation object
        :rtype: Evaluation
        :raises CommandsMapBuilderError: Raised when cannot get evaluation from cache.
        """
        try:
            if self.previous is not None:
                return self.configuration.cache.get_evaluation(self.previous - 1)
            return self.configuration.cache.recent_failed_evaluation
        except CacheError as error:
            raise CommandsMapBuilderError(str(error)) from error

    def get_sources(self) -> List[Path]:
        """
        Get sources list for the commands map.

        :return: Sources list for the commands map
        :rtype: List[Path]
        """
        if self.specified_sources is None or len(self.specified_sources) == 0:
            return self.configuration.sources_repository.sources_list
        return list(self.specified_sources)



"""Add commands to configuration interactively."""
from typing import Dict, List, Optional, Set

import click
import click_params as clickp

from statue.cli.config.interactive_adders.adders_utils import (
    get_context,
    get_contexts,
    get_help_string,
)
from statue.cli.styled_strings import bullet_style, failure_style, name_style
from statue.command_builder import CommandBuilder, ContextSpecification
from statue.config.commands_repository import CommandsRepository
from statue.config.configuration import Configuration
from statue.config.contexts_repository import ContextsRepository
from statue.context import Context
from statue.exceptions import InconsistentConfiguration


class InteractiveCommandAdder:
    """Singleton class for adding and editing commands in configuration instance."""

    @classmethod
    def add_command(cls, configuration: Configuration):
        """
        Add commands interactively to commands repository.

        :param configuration: Statue configuration
        :type configuration: Configuration
        """
        name = cls.get_command_name(configuration.commands_repository)
        help_string = get_help_string(name)
        default_args = cls.get_args(name, args_type="default")
        version = cls.get_version(name)
        preoccupied_contexts = set()
        required_contexts = get_contexts(
            name=name,
            contexts_repository=configuration.contexts_repository,
            name_style_method=name_style,
            contexts_type="required",
        )
        preoccupied_contexts.update(required_contexts)
        allowed_contexts = get_contexts(
            name=name,
            contexts_repository=configuration.contexts_repository,
            name_style_method=name_style,
            contexts_type="allowed",
            preoccupied_contexts=preoccupied_contexts,
        )
        preoccupied_contexts.update(allowed_contexts)
        denied_contexts = get_contexts(
            name=name,
            contexts_repository=configuration.contexts_repository,
            name_style_method=name_style,
            contexts_type="denied",
            preoccupied_contexts=preoccupied_contexts,
        )
        preoccupied_contexts.update(denied_contexts)
        contexts_specifications = cls.get_contexts_specifications(
            name=name,
            contexts_repository=configuration.contexts_repository,
            preoccupied_contexts=preoccupied_contexts,
        )
        configuration.commands_repository.add_command_builders(
            CommandBuilder(
                name=name,
                help=help_string,
                default_args=default_args,
                version=version,
                required_contexts=required_contexts,
                allowed_contexts=allowed_contexts,
                denied_contexts=denied_contexts,
                contexts_specifications=contexts_specifications,
            )
        )

    @classmethod
    def edit_command(cls, name: str, configuration: Configuration):
        """
        Edit a command interactively in commands repository.

        :param name: Name of the command to edit
        :type name: str
        :param configuration: Statue configuration
        :type configuration: Configuration
        """
        command_builder = configuration.commands_repository[name]
        command_builder.help = get_help_string(name)
        command_builder.default_args = cls.get_args(name, args_type="default")
        command_builder.version = cls.get_version(name)
        command_builder.reset_all_contexts()
        preoccupied_contexts = set()
        command_builder.required_contexts = get_contexts(  # type: ignore
            name=name,
            contexts_repository=configuration.contexts_repository,
            name_style_method=name_style,
            contexts_type="required",
        )
        preoccupied_contexts.update(command_builder.required_contexts)
        command_builder.allowed_contexts = get_contexts(  # type: ignore
            name=name,
            contexts_repository=configuration.contexts_repository,
            name_style_method=name_style,
            contexts_type="allowed",
            preoccupied_contexts=preoccupied_contexts,
        )
        preoccupied_contexts.update(command_builder.allowed_contexts)
        command_builder.denied_contexts = get_contexts(  # type: ignore
            name=name,
            contexts_repository=configuration.contexts_repository,
            name_style_method=name_style,
            contexts_type="denied",
            preoccupied_contexts=preoccupied_contexts,
        )
        preoccupied_contexts.update(command_builder.denied_contexts)
        command_builder.contexts_specifications = cls.get_contexts_specifications(
            name=name,
            contexts_repository=configuration.contexts_repository,
            preoccupied_contexts=preoccupied_contexts,
        )

    @classmethod
    def get_command_name(cls, commands_repository: CommandsRepository) -> str:
        """
        Get context name interactively from user.

        :param commands_repository: Commands repository to check
            pre-existing commands in
        :type commands_repository: CommandsRepository
        :return: new command name
        :rtype: str
        """
        name = ""
        while name == "":
            name = click.prompt(
                f"Please choose {bullet_style('command')} name",
                default="",
                show_default=False,
            )
            name = name.strip()
            if name in commands_repository:
                click.echo(failure_style(f"{name} already exists in repository!"))
                name = ""
        return name

    @classmethod
    def get_args(cls, name: str, args_type: Optional[str] = None) -> List[str]:
        """
        Get default arguments for command from user.

        :param name: Name of the command
        :type name: str
        :param args_type: Type of the arguments to get. Optional
        :type args_type: Optional[str]
        :return: List of default arguments
        :rtype: List[str]
        """
        arguments_title = "arguments" if args_type is None else f"{args_type} arguments"
        default_args = click.prompt(
            f"Please add {arguments_title} of {name_style(name)} separated by spaces "
            "(press enter to skip)",
            default="",
            type=clickp.StringListParamType(separator=" "),
            show_default=False,
        )
        default_args = [arg for arg in default_args if arg.strip() != ""]
        return default_args

    @classmethod
    def get_version(cls, name: str) -> Optional[str]:
        """
        Get version string for command from user.

        Press enter to return empty version

        :param name: Name of the command
        :type name: str
        :return: Optional version
        :rtype: Optional[str]
        """
        version = click.prompt(
            f"Please add {name_style(name)} version (press enter to skip)",
            default="",
            type=str,
            show_default=False,
        )
        version = version.strip()
        return None if version == "" else version

    @classmethod
    def get_contexts_specifications(
        cls,
        name: str,
        contexts_repository: ContextsRepository,
        preoccupied_contexts: Set[Context],
    ) -> Dict[Context, ContextSpecification]:
        """
        Get contexts specifications dictionary from user.

        :param name: Name of the command to get contexts specifications for
        :type name: str
        :param contexts_repository: Contexts repository to get contexts from
        :type contexts_repository: ContextsRepository
        :param preoccupied_contexts: Set of already taken contexts
            that can't be specified
        :type preoccupied_contexts: Set[Context]
        :return: Dictionary from contexts to their specifications
        :rtype: Dict[Context, ContextSpecification]
        """
        contexts_specification = {}
        while True:
            context = get_context(
                contexts_repository=contexts_repository,
                name=name,
                name_style_method=name_style,
                context_type="specified",
                preoccupied_contexts=preoccupied_contexts,
            )
            if context is None:
                break
            contexts_specification[context] = cls.get_context_specification(
                context.name
            )
            preoccupied_contexts.add(context)
        return contexts_specification

    @classmethod
    def get_context_specification(cls, name: str) -> ContextSpecification:
        """
        Get context specification for a given context name.

        :param name: Name of the context to be specified
        :type name: str
        :return: Context specification for the given context
        :rtype: ContextSpecification
        """
        while True:
            args = cls.get_args(name, args_type="override")
            added_args = cls.get_args(name, args_type="added")
            clear_args = click.confirm(
                f"Would you like {name_style(name)} to clear arguments?", default=False
            )
            try:
                return ContextSpecification(
                    args=(args if len(args) != 0 else None),
                    add_args=(added_args if len(added_args) != 0 else None),
                    clear_args=clear_args,
                )
            except InconsistentConfiguration as error:
                click.echo(failure_style(str(error)))



"""Context configuration CLI."""
import sys

import click

from statue.cli.common_flags import config_path_option
from statue.cli.config.config_cli import config_cli
from statue.cli.config.interactive_adders.interactive_context_adder import (
    InteractiveContextAdder,
)
from statue.cli.styled_strings import failure_style, name_style
from statue.config.configuration import Configuration
from statue.exceptions import UnknownContext


@config_cli.command("add-context")
@config_path_option
def add_context_cli(config):
    """Add new context to configuration."""
    if config is None:
        config = Configuration.configuration_path()
    configuration = Configuration.from_file(config)
    InteractiveContextAdder.add_context(configuration.contexts_repository)
    configuration.to_toml(config)
    click.echo("Context was successfully added!")


@config_cli.command("edit-context")
@config_path_option
@click.argument("context_name", type=str)
def edit_context_cli(config, context_name):
    """Edit context from configuration."""
    if config is None:
        config = Configuration.configuration_path()
    configuration = Configuration.from_file(config)
    InteractiveContextAdder.edit_context(
        name=context_name, contexts_repository=configuration.contexts_repository
    )
    configuration.to_toml(config)
    click.echo(f"{name_style(context_name)} was successfully edited!")


@config_cli.command("remove-context")
@config_path_option
@click.argument("context_name", type=str)
def remove_context_cli(config, context_name):
    """Remove context from configuration."""
    if config is None:
        config = Configuration.configuration_path()
    configuration = Configuration.from_file(config)
    try:
        context = configuration.contexts_repository[context_name]
    except UnknownContext as error:
        click.echo(failure_style(str(error)))
        sys.exit(1)
    if not click.confirm(
        f"Are you sure you would like to remove the context {name_style(context_name)} "
        "and all of its references from configuration?",
    ):
        click.echo("Abort!")
        return
    configuration.remove_context(context)
    configuration.to_toml(config)
    click.echo(f"{name_style(context_name)} was successfully removed!")



"""Utility module for interactive adders."""
from typing import Callable, List, Optional, Set

import click
import click_params as clickp

from statue.cli.styled_strings import bullet_style, failure_style, name_style
from statue.config.contexts_repository import ContextsRepository
from statue.context import Context
from statue.exceptions import UnknownContext


def get_help_string(name: str) -> str:
    """
    Get help string for an object.

    :param name: Name of the object to get help for
    :type name: str
    :return: help string for the object
    :rtype: str
    """
    help_string = ""
    while help_string == "":
        help_string = click.prompt(
            f"Please add help string for {name_style(name)}",
            default="",
            show_default=False,
        )
        help_string = help_string.strip()
        if help_string == "":
            click.echo(failure_style("Help string cannot be empty!"))
    return help_string


def get_context(
    contexts_repository: ContextsRepository,
    name: str,
    preoccupied_contexts: Set[Context],
    name_style_method: Callable[[str], str],
    context_type: Optional[str] = None,
) -> Optional[Context]:
    """
    Get a context from user for specific configuration (command or source).

    :param contexts_repository: Contexts repository to get contexts from
    :type contexts_repository: ContextsRepository
    :param name: Configuration object to get contexts for
    :type name: str
    :param name_style_method: Styling method for name.
    :type name_style_method: Callable[[str], str]
    :param context_type: Optional type of the desired contexts.
    :type context_type: Optional[str]
    :param preoccupied_contexts: Contexts that cannot be set by the users.
    :type preoccupied_contexts: Set[Context]
    :return: Contexts list for given item
    :rtype: List[Context]
    """
    if len(contexts_repository) == 0:
        return None
    contexts_options = ", ".join(
        [
            name_style(context.name)
            for context in contexts_repository
            if context not in preoccupied_contexts
        ]
    )
    name = name_style_method(name)
    context_title = "context" if context_type is None else f"{context_type} context"
    while True:
        try:
            context_name = click.prompt(
                f"Add {bullet_style(context_title)} to {name} "
                f"(options: [{contexts_options}], press enter to skip)",
                default="",
                type=str,
                show_default=False,
            ).strip()
            if context_name == "":
                return None
            context = contexts_repository[context_name]
            if context in preoccupied_contexts:
                click.echo(
                    failure_style(
                        f"Could not set {context_name} as specifies context "
                        "because it is preoccupied."
                    )
                )
                continue
            return context
        except UnknownContext as error:
            click.echo(failure_style(str(error)))


def get_contexts(
    contexts_repository: ContextsRepository,
    name: str,
    name_style_method: Callable[[str], str],
    contexts_type: Optional[str] = None,
    preoccupied_contexts: Optional[Set[Context]] = None,
) -> List[Context]:
    """
    Get contexts from user for specific configuration (command or source).

    :param contexts_repository: Contexts repository to get contexts from
    :type contexts_repository: ContextsRepository
    :param name: Configuration object to get contexts for
    :type name: str
    :param name_style_method: Styling method for name.
    :type name_style_method: Callable[[str], str]
    :param contexts_type: Optional type of the desired contexts.
    :type contexts_type: Optional[str]
    :param preoccupied_contexts: Contexts that cannot be set by the users. Optional.
    :type preoccupied_contexts: Optional[Set[Context]]:
    :return: Contexts list for given item
    :rtype: List[Context]
    """
    if len(contexts_repository) == 0:
        return []
    preoccupied_contexts = (
        set() if preoccupied_contexts is None else preoccupied_contexts
    )
    contexts_options = ", ".join(
        [
            name_style(context.name)
            for context in contexts_repository
            if context not in preoccupied_contexts
        ]
    )
    name = name_style_method(name)
    contexts_title = (
        "contexts" if contexts_type is None else f"{contexts_type} contexts"
    )
    while True:
        try:
            context_names = click.prompt(
                f"Add {bullet_style(contexts_title)} to {name} "
                f"(options: [{contexts_options}])",
                default="",
                type=clickp.StringListParamType(),
                show_default=False,
            )
            context_names = [
                context_name.strip()
                for context_name in context_names
                if context_name.strip() != ""
            ]
            contexts = [contexts_repository[context] for context in context_names]
            unavailable_contexts = [
                context.name for context in contexts if context in preoccupied_contexts
            ]
            if len(unavailable_contexts) > 0:
                unavailable_contexts_string = ", ".join(
                    context_name for context_name in unavailable_contexts
                )
                click.echo(
                    failure_style(
                        "The following contexts could not be set: "
                        f"{unavailable_contexts_string}"
                    )
                )
                continue
            return contexts
        except UnknownContext as error:
            click.echo(failure_style(str(error)))



"""Add sources to configuration interactively."""
from pathlib import Path
from typing import Callable, List, Optional

import click
import click_params as clickp
import git

from statue.cli.config.interactive_adders.adders_utils import get_contexts
from statue.cli.styled_strings import (
    bullet_style,
    failure_style,
    name_style,
    source_style,
    success_style,
)
from statue.commands_filter import CommandsFilter
from statue.config.configuration import Configuration
from statue.sources_finder import expend

YES = ["y", "yes"]
NO = ["n", "no"]
EXPEND = ["e", "expend"]


class InteractiveSourcesAdder:
    """Singleton class for adding sources to configuration instance."""

    @classmethod
    def update_sources_repository(
        cls,
        configuration: Configuration,
        sources: List[Path],
        repo: Optional[git.Repo] = None,
        exclude: Optional[List[Path]] = None,
    ):
        """
        Add sources to configuration instance.

        :param configuration: The configuration to add sources to
        :type configuration: Configuration
        :param sources: Sources list to add to the configuration
        :type sources: List[Path]
        :param repo: Optional repository instance for avoiding ignored files.
        :type repo: Optional[Repo]
        :param exclude: Optional. Paths to be excluded from tracking
        :type exclude: Optional[List[Path]]
        """
        for source in sources:
            choices = cls._get_choices(configuration=configuration, source=source)
            if choices == NO:
                click.echo(
                    failure_style(
                        f"{source} was already defined in configuration. Skipping..."
                    )
                )
                continue
            choices_string = cls._get_choices_string(choices)
            option = click.prompt(
                f"Would you like to track {source_style(str(source))} "
                f"({choices_string}. default: {success_style('yes')})",
                type=click.Choice(choices, case_sensitive=False),
                show_choices=False,
                show_default=False,
                default="yes",
            ).lower()
            if option in EXPEND:
                cls.update_sources_repository(
                    configuration,
                    expend(source, repo=repo, exclude=exclude),
                    repo=repo,
                    exclude=exclude,
                )
            if option not in YES:
                continue
            configuration.sources_repository[source] = cls.get_filter(
                configuration, source
            )

    @classmethod
    def get_filter(cls, configuration: Configuration, source: Path) -> CommandsFilter:
        """
        Get a commands filter from the user for the specified source.

        :param configuration: Configuration instance
        :type configuration: Configuration
        :param source: Source to get filter for
        :type source: Path
        :return: Commands filter for the source
        :rtype: CommandsFilter
        """
        while True:
            try:
                contexts = get_contexts(
                    contexts_repository=configuration.contexts_repository,
                    name=str(source),
                    name_style_method=source_style,
                )
                allowed_commands = cls.get_commands(
                    configuration, source, prefix="allowed", style_func=success_style
                )
                denied_commands = cls.get_commands(
                    configuration, source, prefix="denied", style_func=failure_style
                )
                return CommandsFilter(
                    contexts=contexts,
                    allowed_commands=allowed_commands,
                    denied_commands=denied_commands,
                )
            except ValueError as error:
                click.echo(failure_style(str(error)))
                click.echo("Try again...")

    @classmethod
    def get_commands(
        cls,
        configuration: Configuration,
        source: Path,
        prefix: str,
        style_func: Callable[[str], str],
    ) -> Optional[List[str]]:
        """
        Get commands list from user for specific source.

        :param configuration: Configuration instance
        :type configuration: Configuration
        :param source: Source to get commands for
        :type source: Path
        :param prefix: Commands type to look for (allowed/denied)
        :type prefix: str
        :param style_func: Styling function for prefix
        :type style_func: Callable[[str], str]
        :return: Sources list for source
        :rtype: List[Context]
        """
        if len(configuration.commands_repository) == 0:
            return None
        commands_options = ", ".join(
            [
                name_style(command_builder.name)
                for command_builder in configuration.commands_repository
            ]
        )
        while True:
            command_names = click.prompt(
                f"Add {style_func(prefix)} commands to {source_style(str(source))} "
                f"(options: [{commands_options}], "
                f"No specified {prefix} commands by default)",
                default="",
                type=clickp.StringListParamType(),
                show_default=False,
            )
            command_names = [
                command_name.strip()
                for command_name in command_names
                if command_name.strip() != ""
            ]
            if len(command_names) == 0:
                return None
            unknown_commands = [
                command_name
                for command_name in command_names
                if command_name not in configuration.commands_repository
            ]
            if len(unknown_commands) != 0:
                click.echo(
                    failure_style(
                        "Could not find the following commands: "
                        f"{', '.join(unknown_commands)}"
                    )
                )
                continue
            return command_names

    @classmethod
    def _get_choices(cls, configuration: Configuration, source: Path):
        choices = []
        if source not in configuration.sources_repository.sources_list:
            choices.extend(YES)
        choices.extend(NO)
        if source.is_dir():
            choices.extend(EXPEND)
        return choices

    @classmethod
    def _get_choices_string(cls, choices):
        choices_string = f"{success_style('[Y]es')}, {failure_style('[N]o')}"
        if EXPEND[0] in choices:
            choices_string += f", {bullet_style('[E]xpend')}"
        return choices_string



"""Find all python sources in a directory."""
from pathlib import Path
from typing import List, Optional

from git import Repo

from statue.io_util import is_equal_or_child_of


def find_sources(
    path: Path, repo: Repo = None, exclude: Optional[List[Path]] = None
) -> List[Path]:
    """
    Search for sources recursively.

    :param path: Path to a dictionary or file
    :type path: Path
    :param repo: Optional. A repository instance. Used to find ignored files
    :type repo: Repo
    :param exclude: Optional. List of sources to exclude
    :type exclude: Optional[List[Path]]
    :return: List of sources
    :rtype: List[Path]
    """
    if exclude is None:
        exclude = []
    if any(is_equal_or_child_of(path, excluded_source) for excluded_source in exclude):
        return []
    if is_python(path):
        return [path]
    if not path.is_dir():
        return []
    return expend(path, repo=repo, exclude=exclude)


def expend(
    path: Path, repo: Repo = None, exclude: Optional[List[Path]] = None
) -> List[Path]:
    """
    Find all sources inside a directory which are not ignored.

    :param path: Path to a dictionary or file
    :type path: Path
    :param repo: Optional. A repository instance. Used to find ignored files
    :type repo: Repo
    :param exclude: Optional. List of sources to exclude
    :type exclude: Optional[List[Path]]
    :return: List of sources
    :rtype: List[Path]
    """
    if exclude is None:
        exclude = []
    inner_files = [
        inner_path
        for inner_path in path.iterdir()
        if all(
            not is_equal_or_child_of(inner_path, excluded_source)
            for excluded_source in exclude
        )
    ]
    if repo is not None:
        ignored_files = [Path(ignored) for ignored in repo.ignored(*inner_files)]
        inner_files = [
            inner_path for inner_path in inner_files if inner_path not in ignored_files
        ]
    sources = []
    for inner_path in inner_files:
        sources.extend(find_sources(inner_path, repo=repo, exclude=exclude))
    return sorted(sources)


def is_python(path: Path) -> bool:
    """
    Is path a python module or package.

    :param path: Path to a dictionary or file
    :type path: Path
    :return: Is python module or package
    :rtype: bool
    """
    return is_python_module(path) or is_python_package(path)


def is_python_module(path: Path) -> bool:
    """
    Is given path a python module.

    :param path: Path to a dictionary or file
    :type path: Path
    :return: Is python module
    :rtype: bool
    """
    return path.is_file() and path.suffix == ".py"


def is_python_package(path: Path) -> bool:
    """
    Is path a python package.

    :param path: Path to a dictionary or file
    :type path: Path
    :return: Is python package
    :rtype: bool
    """
    return path.is_dir() and (path / "__init__.py").exists()



# noqa: D100
# pylint: disable=missing-module-docstring
import asyncio
import os
import subprocess  # nosec
import time
from dataclasses import asdict, dataclass, field
from pathlib import Path
from typing import Any, Dict, List

from statue.constants import ENCODING
from statue.exceptions import CommandExecutionError
from statue.sources_locks_repository import SourcesLocksRepository


@dataclass
class CommandEvaluation:
    """Evaluation result of a command."""

    command: "Command"
    success: bool
    execution_duration: float
    captured_output: List[str] = field(default_factory=list)

    @property
    def captured_output_string(self):
        """Join captured output into a single string."""
        return "\n".join(self.captured_output)

    def as_dict(self) -> Dict[str, Any]:
        """
        Return command evaluation as json dictionary.

        :return: Self as dictionary
        :rtype: Dict[str, Any]
        """
        command_json = {
            key: value
            for key, value in asdict(self.command).items()
            if value is not None
        }
        return dict(
            command=command_json,
            execution_duration=self.execution_duration,
            captured_output=self.captured_output,
            success=self.success,
        )

    @classmethod
    def from_dict(cls, command_evaluation: Dict[str, Any]) -> "CommandEvaluation":
        """
        Read command evaluation from json dictionary.

        :param command_evaluation: Json command evaluation
        :type command_evaluation: Dict[str, Any]
        :return: Parsed command evaluation
        :rtype: CommandEvaluation
        """
        return CommandEvaluation(
            command=Command(**command_evaluation["command"]),
            success=command_evaluation["success"],
            execution_duration=command_evaluation["execution_duration"],
            captured_output=command_evaluation["captured_output"],
        )


@dataclass
class Command:
    """Runnable evaluation command."""

    name: str
    args: List[str] = field(default_factory=list)

    def program_execution_args(self, source: Path) -> List[str]:
        """
        Get the program command to be run as a subprocess.

        :param source: The source to run the command on.
        :type source: Path
        :return: Program arguments list
        :rtype: List[str]
        """
        return [self.name, str(source), *self.args]

    def execute(self, source: Path) -> CommandEvaluation:
        """
        Execute the command.

        :param source: source files to check.
        :type: str
        :return: Command's evaluation including the command itself and is it successful
        :rtype: CommandEvaluation
        :raises CommandExecutionError: raised when command is not found.
        """
        try:
            start_time = time.time()
            subprocess_result = subprocess.run(
                self.program_execution_args(source),
                env=os.environ,
                check=False,
                capture_output=True,
            )
            end_time = time.time()
        except FileNotFoundError as error:
            raise CommandExecutionError(self.name) from error
        captured_stdout = subprocess_result.stdout.decode(ENCODING)
        captured_stderr = subprocess_result.stderr.decode(ENCODING)
        return CommandEvaluation(
            command=self,
            success=(subprocess_result.returncode == 0),
            execution_duration=end_time - start_time,
            captured_output=self._build_captured_output(
                captured_stdout=captured_stdout, captured_stderr=captured_stderr
            ),
        )

    async def execute_async(self, source: Path) -> CommandEvaluation:
        """
        Execute the command asynchronously.

        :param source: source files to check.
        :type source: Path
        :return: Command's evaluation including the command itself and is it successful
        :rtype: CommandEvaluation
        :raises CommandExecutionError: raised when command is not found.
        """
        source_lock = await SourcesLocksRepository.get_lock(source)
        try:
            await source_lock.acquire()
            start_time = time.time()
            async_process = await asyncio.create_subprocess_exec(
                *self.program_execution_args(source),
                stdout=asyncio.subprocess.PIPE,
                stderr=asyncio.subprocess.PIPE,
                env=os.environ,
            )
            stdout, stderr = await async_process.communicate()
            end_time = time.time()
            source_lock.release()
        except FileNotFoundError as error:
            raise CommandExecutionError(self.name) from error
        captured_stdout, captured_stderr = stdout.decode(ENCODING), stderr.decode(
            ENCODING
        )
        return CommandEvaluation(
            command=self,
            success=(async_process.returncode == 0),
            execution_duration=end_time - start_time,
            captured_output=self._build_captured_output(
                captured_stdout=captured_stdout, captured_stderr=captured_stderr
            ),
        )

    @classmethod
    def _build_captured_output(cls, captured_stdout, captured_stderr):
        captured_output_as_string = captured_stdout + captured_stderr
        return (
            captured_output_as_string.split("\n")
            if len(captured_output_as_string) != 0
            else []
        )



"""Main CLI for statue."""
from pathlib import Path
from typing import Optional

import click

from statue import __version__
from statue.cli.common_flags import config_path_option
from statue.cli.styled_strings import failure_style
from statue.config.configuration import Configuration
from statue.exceptions import StatueConfigurationError

pass_configuration = click.make_pass_decorator(Configuration)


@click.group(name="statue", no_args_is_help=True)
@click.version_option(version=__version__)
@config_path_option
@click.option(
    "--cache-dir",
    envvar="STATUE_CACHE",
    type=click.Path(exists=True, file_okay=False, path_type=Path),
    help="Statue caching directory path",
)
@click.pass_context
def statue_cli(
    ctx: click.Context,
    config: Optional[Path],
    cache_dir: Optional[Path],
):
    """Statue is a static code analysis tools orchestrator."""
    if ctx.invoked_subcommand in ["config", "templates"]:
        return
    try:
        ctx.obj = Configuration.from_file(config_path=config, cache_dir=cache_dir)
    except StatueConfigurationError as error:
        click.echo(failure_style(click.style(error)))
        ctx.exit(3)



"""Sources configuration CLI."""
import sys
from pathlib import Path
from typing import List, Optional

import click

from statue.cli.common_flags import config_path_option
from statue.cli.config.config_cli import config_cli
from statue.cli.config.interactive_adders.interactive_sources_adder import (
    InteractiveSourcesAdder,
)
from statue.cli.styled_strings import failure_style, source_style
from statue.config.configuration import Configuration


@config_cli.command("add-source")
@click.argument("sources", type=click.Path(exists=True, path_type=Path), nargs=-1)
@config_path_option
def add_source_to_configuration_cli(config: Optional[Path], sources: List[Path]):
    """Add new source to configuration."""
    if config is None:
        config = Configuration.configuration_path()
    configuration = Configuration.from_file(config)
    InteractiveSourcesAdder.update_sources_repository(configuration, list(sources))
    configuration.to_toml(config)
    click.echo("Sources were added successfully!")


@config_cli.command("edit-source")
@click.argument("source", type=click.Path(exists=True, path_type=Path))
@config_path_option
def edit_source_in_configuration_cli(config: Optional[Path], source: Path):
    """Edit a source from configuration."""
    if config is None:
        config = Configuration.configuration_path()
    configuration = Configuration.from_file(config)
    if source not in configuration.sources_repository.sources_list:
        click.echo(failure_style(f"{source} is not specified in configuration"))
        sys.exit(1)
    configuration.sources_repository[source] = InteractiveSourcesAdder.get_filter(
        configuration=configuration, source=source
    )


@config_cli.command("remove-source")
@click.argument("source", type=click.Path(exists=True, path_type=Path))
@config_path_option
def remove_source_from_configuration_cli(config: Optional[Path], source: Path):
    """Remove context from configuration."""
    if config is None:
        config = Configuration.configuration_path()
    configuration = Configuration.from_file(config)
    if source not in configuration.sources_repository.sources_list:
        click.echo(failure_style(f"Could not find {source} in configuration."))
        sys.exit(1)
    if not click.confirm(
        f"Are you sure you would like to remove the source {source_style(str(source))} "
        "from configuration?",
    ):
        click.echo("Abort!")
        return
    configuration.sources_repository.remove_source(source)
    configuration.to_toml(config)
    click.echo(f"{source_style(str(source))} was successfully removed!")



"""Initialize configuration CLI."""
import sys
from pathlib import Path
from typing import List, Optional

import click
import git

from statue.cli.common_flags import config_path_option
from statue.cli.config.config_cli import config_cli
from statue.cli.config.interactive_adders.interactive_sources_adder import (
    InteractiveSourcesAdder,
)
from statue.cli.styled_strings import failure_style
from statue.config.configuration import Configuration
from statue.exceptions import StatueConfigurationError, UnknownTemplate
from statue.sources_finder import find_sources
from statue.templates.templates_provider import TemplatesProvider


@config_cli.command("init", short_help="Initialize configuration for Statue.")
@config_path_option
@click.option(
    "-t",
    "--template",
    type=str,
    default="defaults",
    help="Configuration template name",
)
@click.option(
    "--with-sources/--no-sources",
    is_flag=True,
    default=True,
    help="Track available sources and add to configuration",
)
@click.option(
    "-e",
    "--exclude",
    type=click.Path(exists=True, path_type=Path),
    multiple=True,
    help="Sources paths that Statue will ignore tracking",
)
@click.option("-y", "interactive", flag_value=False, default=True)
@click.option(
    "--git/--no-git",
    "use_git",
    is_flag=True,
    default=True,
    help="Should use git to trace ignored files",
)
@click.option(
    "--fix-versions",
    is_flag=True,
    default=False,
    help="Fix versions in config when running",
)
@click.option(
    "-i",
    "--install",
    is_flag=True,
    default=False,
    help="Install latest version for all commands in configuration",
)
@click.option(
    "--blank", is_flag=True, default=False, help="Initialize blank configuration"
)
def init_config_cli(  # pylint: disable=too-many-arguments
    config: Optional[Path],
    template: str,
    with_sources: bool,
    exclude: List[Path],
    interactive: bool,
    use_git: bool,
    fix_versions: bool,
    install: bool,
    blank: bool,
):
    """
    Initialize configuration for Statue.

    By default, this command searches for sources files in the given directory (cwd by
     default) and sets them with default contexts.

    You can run this command with the "-i" flag in order to choose interactively which
     source files to track and which contexts to assign to them.
    """
    output_path = config if config is not None else Configuration.configuration_path()
    try:
        configuration = _load_configuration(
            template=template,
            exclude=exclude,
            blank=blank,
            with_sources=with_sources,
            use_git=use_git,
            interactive=interactive,
            install=install,
            fix_versions=fix_versions,
        )
    except (UnknownTemplate, StatueConfigurationError) as error:
        click.echo(failure_style(str(error)))
        sys.exit(3)
    configuration.to_toml(output_path)
    click.echo("Done!")


def _load_configuration(  # pylint: disable=too-many-arguments
    template: str,
    exclude: List[Path],
    blank: bool,
    with_sources: bool,
    use_git: bool,
    interactive: bool,
    install: bool,
    fix_versions: bool,
):
    if blank:
        return Configuration.empty_configuration()
    configuration = Configuration.from_file(
        TemplatesProvider.get_template_path(template)
    )
    if with_sources:
        _update_sources(
            configuration=configuration,
            use_git=use_git,
            interactive=interactive,
            exclude=exclude,
        )
    if fix_versions or install:
        for command_builder in configuration.commands_repository:
            if install:
                command_builder.update_to_version()
            if fix_versions and command_builder.installed():
                command_builder.set_version_as_installed()
    return configuration


def _update_sources(
    configuration: Configuration, exclude: List[Path], use_git: bool, interactive: bool
):
    directory = Path.cwd()
    repo = None
    if use_git:
        try:
            repo = git.Repo(directory)
        except git.InvalidGitRepositoryError:
            pass
    sources = [
        source.relative_to(directory)
        for source in find_sources(directory, repo=repo, exclude=exclude)
    ]
    if interactive:
        InteractiveSourcesAdder.update_sources_repository(
            configuration=configuration, sources=sources, repo=repo, exclude=exclude
        )
        return
    configuration.sources_repository.track_sources(*sources)



"""Filter commands using simple checks."""
from collections import OrderedDict
from typing import Any, Collection, FrozenSet, List, Optional
from typing import OrderedDict as OrderedDictType

from statue.command_builder import CommandBuilder
from statue.constants import ALLOW_LIST, CONTEXTS, DENY_LIST
from statue.context import Context


class CommandsFilter:
    """Filter data class for filtering command builders."""

    def __init__(
        self,
        contexts: Optional[List[Context]] = None,
        allowed_commands: Optional[Collection[str]] = None,
        denied_commands: Optional[Collection[str]] = None,
    ):
        """
        Initialize commands filter.

        :param contexts: Filter contexts
        :type contexts: Optional[Collection[Context]]
        :param allowed_commands: Allowed commands that pass filter
        :type allowed_commands: Optional[Collection[str]]
        :param denied_commands: Denied commands that do not pass filter
        :type denied_commands: Optional[Collection[str]]
        :raises ValueError: Raised if both allowed and denied commands are set.
        """
        self._contexts = [] if contexts is None else list(contexts)
        self._allowed_commands = (
            None if allowed_commands is None else frozenset(allowed_commands)
        )
        self._denied_commands = (
            None if denied_commands is None else frozenset(denied_commands)
        )
        if self.allowed_commands is not None and self.denied_commands is not None:
            raise ValueError(
                "Commands filter cannot be set with both allowed and denied commands"
            )

    @property
    def contexts(self) -> List[Context]:
        """Filter contexts."""
        return list(self._contexts)

    @property
    def allowed_commands(self) -> Optional[FrozenSet[str]]:
        """Filter allowed commands."""
        return self._allowed_commands

    @property
    def denied_commands(self) -> Optional[FrozenSet[str]]:
        """Filter denied commands."""
        return self._denied_commands

    def __eq__(self, other: object) -> bool:
        """
        Compare command filters.

        :param other: Other object to compare to
        :type other: object
        :return: is equal
        :rtype: bool
        """
        return (
            isinstance(other, CommandsFilter)
            and self.contexts == other.contexts
            and self.allowed_commands == other.allowed_commands
            and self.denied_commands == other.denied_commands
        )

    def __str__(self) -> str:
        """
        Self as string.

        :return: Self as a representation string
        :rtype: str
        """
        return repr(self)

    def __repr__(self) -> str:
        """
        Representation string.

        :return: Self as a representation string
        :rtype: str
        """
        return (
            "CommandsFilter("
            f"contexts={self.contexts}, "
            f"allowed_commands={self.allowed_commands}, "
            f"denied_commands={self.denied_commands})"
        )

    def pass_filter(self, command_builder: CommandBuilder) -> bool:
        """
        Check if given command builder pass this filter.

        :param command_builder: Commands builder to be checked
        :type command_builder: CommandBuilder
        :return: Does this command builder pass the filter
        :rtype: bool
        """
        if (
            self.allowed_commands is not None
            and command_builder.name not in self.allowed_commands
        ):
            return False
        if (
            self.denied_commands is not None
            and command_builder.name in self.denied_commands
        ):
            return False
        return command_builder.match_contexts(*self.contexts)

    def as_dict(self) -> OrderedDictType[str, List[str]]:
        """
        Encode commands filter as a dictionary.

        This is used in order to serialize the commands filter in
        a configuration file.

        :return: Serialized representation dictionary
        :rtype: OrderedDict[str, List[str]]
        """
        filter_as_dict = OrderedDict()
        if len(self.contexts) != 0:
            context_names = [context.name for context in self.contexts]
            context_names.sort()
            filter_as_dict[CONTEXTS] = context_names
        if self.allowed_commands is not None:
            allowed_commands = list(self.allowed_commands)
            allowed_commands.sort()
            filter_as_dict[ALLOW_LIST] = allowed_commands
        if self.denied_commands is not None:
            denied_commands = list(self.denied_commands)
            denied_commands.sort()
            filter_as_dict[DENY_LIST] = denied_commands
        return filter_as_dict

    @classmethod
    def merge(
        cls, filter1: "CommandsFilter", filter2: "CommandsFilter"
    ) -> "CommandsFilter":
        """
        Merge two command filters into one filter.

        PAY ATTENTION: the order of merge is important! evaluating contexts will be done
        in their order

        :param filter1: First filter to be merged
        :type filter1: CommandsFilter
        :param filter2: Second filter to be merged
        :type filter2: CommandsFilter
        :return: Merged filter
        :rtype: CommandsFilter
        :raises ValueError: Raised when a command is allowed in one filter
            and denied in another
        """
        contexts = filter1.contexts + filter2.contexts
        allowed_commands = cls._intersect_optional_sets(
            filter1.allowed_commands, filter2.allowed_commands
        )
        denied_commands = cls._union_optional_sets(
            filter1.denied_commands, filter2.denied_commands
        )
        if allowed_commands is not None and denied_commands is not None:
            both_allowed_and_denied = allowed_commands.intersection(denied_commands)
            if len(both_allowed_and_denied) != 0:
                raise ValueError(
                    "Cannot merge command filters because the following commands "
                    "are both allowed and denied: "
                    f"{', '.join(both_allowed_and_denied)}"
                )
            denied_commands = None
        return CommandsFilter(
            contexts=contexts,
            allowed_commands=allowed_commands,
            denied_commands=denied_commands,
        )

    @classmethod
    def _union_optional_sets(
        cls, set1: Optional[FrozenSet[Any]], set2: Optional[FrozenSet[Any]]
    ) -> Optional[FrozenSet[Any]]:
        if set1 is None and set2 is None:
            return None
        set1 = frozenset() if set1 is None else set1
        set2 = frozenset() if set2 is None else set2
        return set1.union(set2)

    @classmethod
    def _intersect_optional_sets(
        cls, set1: Optional[FrozenSet[Any]], set2: Optional[FrozenSet[Any]]
    ) -> Optional[FrozenSet[Any]]:
        if set1 is None:
            return set2
        if set2 is None:
            return set1
        return set1.intersection(set2)



"""Build commands from configuration."""
# pylint: disable=too-many-public-methods,too-many-arguments
# pylint: disable=too-many-instance-attributes
import importlib
import itertools
import os
import subprocess  # nosec
import sys
from collections import OrderedDict
from typing import Any, Dict, Iterable, List, Optional
from typing import OrderedDict as OrderedDictType
from typing import Set

import pkg_resources

from statue.command import Command
from statue.config.contexts_repository import ContextsRepository
from statue.constants import (
    ALLOWED_CONTEXTS,
    ARGS,
    DENIED_CONTEXTS,
    HELP,
    REQUIRED_CONTEXTS,
    VERSION,
)
from statue.context import Context
from statue.context_specification import ContextSpecification
from statue.exceptions import (
    InconsistentConfiguration,
    InvalidCommand,
    InvalidConfiguration,
    MissingHelpString,
    StatueConfigurationError,
    UnknownContext,
)
from statue.verbosity import DEFAULT_VERBOSITY, is_silent


class CommandBuilder:
    """Command builder as specified in configuration."""

    def __init__(
        self,
        name: str,
        help: str,  # pylint: disable=redefined-builtin
        default_args: Optional[List[str]] = None,
        version: Optional[str] = None,
        required_contexts: Optional[Iterable[Context]] = None,
        allowed_contexts: Optional[Iterable[Context]] = None,
        denied_contexts: Optional[Iterable[Context]] = None,
        contexts_specifications: Optional[Dict[Context, ContextSpecification]] = None,
    ):
        """
        Constructor.

        :param name: Name of the command to be built by the builder
        :type name str
        :param help: Help string to describe the command
        :type help: str
        :param default_args: Optional default arguments to be added to the command
        :type default_args: Optional[List[str]]
        :param version: Optional version specification for the command builder
        :type version: Optional[str]
        :param required_contexts: Optional list of contexts required by
            the command builder
        :type required_contexts: Optional[List[Context]]
        :param allowed_contexts: Optional list of contexts allowed for
            the command builder
        :type allowed_contexts: Optional[List[Context]]
        :param denied_contexts: Optional list of contexts denied for
            the command builder
        :type denied_contexts: Optional[List[Context]]
        :param contexts_specifications: Optional dictionary of contexts specification
            for the command builder
        :type contexts_specifications: Optional[Dict[Context, ContextSpecification]]
        """
        self.name = name
        self.help = help
        self.default_args = default_args if default_args is not None else []
        self.version = version

        self._initialize_contexts()
        self.required_contexts = (
            set(required_contexts) if required_contexts is not None else set()
        )
        self.allowed_contexts = (
            set(allowed_contexts) if allowed_contexts is not None else set()
        )
        self.denied_contexts = (
            set(denied_contexts) if denied_contexts is not None else set()
        )
        self.contexts_specifications = (
            contexts_specifications if contexts_specifications is not None else {}
        )

    @property
    def required_contexts(self) -> Set[Context]:
        """Get contexts required by command builder."""
        return self._required_contexts

    @required_contexts.setter
    def required_contexts(self, required_contexts):
        """
        Set contexts required by command builder.

        :param required_contexts: Required contexts to be set.
        :type required_contexts: Set[Context]
        """
        required_contexts = set(required_contexts)
        if len(required_contexts) != 0:
            self._validate_consistency(
                allowed=self.allowed_contexts,
                denied=self.denied_contexts,
                required=required_contexts,
                specified=self.specified_contexts,
            )
        self._required_contexts = required_contexts

    @property
    def allowed_contexts(self) -> Set[Context]:
        """Get contexts allowed for command builder."""
        return self._allowed_contexts

    @allowed_contexts.setter
    def allowed_contexts(self, allowed_contexts):
        """
        Set contexts allowed for command builder.

        :param allowed_contexts: Allowed contexts to be set
        :type allowed_contexts: Set[Context]
        """
        allowed_contexts = set(allowed_contexts)
        if len(allowed_contexts) != 0:
            self._validate_consistency(
                allowed=allowed_contexts,
                denied=self.denied_contexts,
                required=self.required_contexts,
                specified=self.specified_contexts,
            )
        self._allowed_contexts = allowed_contexts

    @property
    def denied_contexts(self) -> Set[Context]:
        """Get contexts allowed for command builder."""
        return self._denied_contexts

    @denied_contexts.setter
    def denied_contexts(self, denied_contexts):
        denied_contexts = set(denied_contexts)
        if len(denied_contexts) != 0:
            self._validate_consistency(
                allowed=self.allowed_contexts,
                denied=denied_contexts,
                required=self.required_contexts,
                specified=self.specified_contexts,
            )
        self._denied_contexts = denied_contexts

    @property
    def contexts_specifications(self) -> Dict[Context, ContextSpecification]:
        """Get contexts specification dictionary for command builder."""
        return self._contexts_specifications

    @contexts_specifications.setter
    def contexts_specifications(self, contexts_specifications):
        """
        Set contexts specification dictionary for command builder.

        :param contexts_specifications: Contexts specification dictionary to be set
        :type contexts_specifications: Dict[Context, ContextSpecification]
        """
        if len(contexts_specifications) != 0:
            self._validate_consistency(
                allowed=self.allowed_contexts,
                denied=self.denied_contexts,
                required=self.required_contexts,
                specified=set(contexts_specifications.keys()),
            )
        self._contexts_specifications = contexts_specifications

    def __repr__(self) -> str:
        """
        Represent context as string.

        :return: String representation of command builder
        :rtype: str
        """
        required_contexts = [context.name for context in self.required_contexts]
        required_contexts.sort()
        allowed_contexts = [context.name for context in self.allowed_contexts]
        allowed_contexts.sort()
        denied_contexts = [context.name for context in self.denied_contexts]
        denied_contexts.sort()
        contexts_specification = {
            context.name: specification
            for context, specification in self.contexts_specifications.items()
        }
        return (
            "CommandBuilder("
            f"name={self.name}, "
            f"help={self.help}, "
            f"default_args={self.default_args}, "
            f"version={self.version}, "
            f"required_contexts={required_contexts}, "
            f"allowed_contexts={allowed_contexts}, "
            f"denied_contexts={denied_contexts}, "
            f"contexts_specifications={contexts_specification}"
            ")"
        )

    def __eq__(self, other: object) -> bool:
        """
        Check equality of this command builder with other object.

        :param other: Object to compare to
        :type other: object
        :return: is equal to self
        :rtype: bool
        """
        return (
            isinstance(other, CommandBuilder)
            and self.name == other.name
            and self.help == other.help
            and self.default_args == other.default_args
            and self.version == other.version
            and self.allowed_contexts == other.allowed_contexts
            and self.denied_contexts == other.denied_contexts
            and self.required_contexts == other.required_contexts
            and self.contexts_specifications == other.contexts_specifications
        )

    @property
    def install_name(self) -> str:
        """
        Name to state while installing with pip.

        When installing a specific version with pip, one should add "==" with the
        specific version afterwards.

        If no version is specified, same as name.

        :return: name and version
        :rtype: str
        """
        if self.version is None:
            return self.name
        return f"{self.name}=={self.version}"

    @property
    def installed_version(self) -> Optional[str]:
        """
        Version of the installed package.

        Might not be the same as the version attribute.

        :return: version of installed package
        :rtype: str or None
        """
        package = self._get_package()
        if package is None:
            return None
        return package.version

    @property
    def specified_contexts(self) -> Set[Context]:
        """Contexts names list with arguments specifications."""
        return set(self.contexts_specifications.keys())

    @property
    def available_contexts(self) -> Set[Context]:
        """Contexts which are available to use according to this command."""
        return {
            *self.required_contexts,
            *self.allowed_contexts,
            *self.specified_contexts,
        }

    def reset_all_contexts(self):
        """Reset all available contexts."""
        self.required_contexts = []
        self.allowed_contexts = []
        self.denied_contexts = []
        self.contexts_specifications = {}

    def installed(self) -> bool:
        """
        Is this command installed.

        :return: Either the command is installed or not
        :rtype: bool
        """
        return self.installed_version is not None

    def set_version_as_installed(self):
        """
        Set version to be the installed version.

        PAY ATTENTION: if command is not installed, version will be set to None
        """
        self.version = self.installed_version

    def installed_correctly(self) -> bool:
        """
        Checks that command is installed and its version matches.

        :return: whether the command is installed correctly
        :rtype: bool
        """
        return self.installed() and self.installed_version_match()

    def installed_version_match(self) -> bool:
        """
        Is the installed version match the specified version.

        :return: is the installed version matches the desired version
        :rtype: bool
        """
        if self.version is None:
            return True
        return self.installed_version == self.version

    def install(self, verbosity: str = DEFAULT_VERBOSITY) -> None:
        """
        Install command using pip.

        :param verbosity: Verbosity level.
        :type verbosity: str
        """
        if self.installed():
            return
        if not is_silent(verbosity):
            print(f"Installing {self.install_name}")
        subprocess.run(  # nosec
            [sys.executable, "-m", "pip", "install", self.install_name],
            env=os.environ,
            check=False,
            capture_output=is_silent(verbosity),
        )

    def update(self, verbosity: str = DEFAULT_VERBOSITY) -> None:
        """
        Update command using pip.

        :param verbosity: Verbosity level.
        :type verbosity: str
        """
        if not is_silent(verbosity):
            print(f"Updating {self.name}")
        subprocess.run(  # nosec
            [sys.executable, "-m", "pip", "install", "-U", self.name],
            env=os.environ,
            check=False,
            capture_output=is_silent(verbosity),
        )

    def uninstall(self, verbosity: str = DEFAULT_VERBOSITY) -> None:
        """
        Uninstall command using pip.

        :param verbosity: Verbosity level.
        :type verbosity: str
        """
        if not self.installed():
            return
        if not is_silent(verbosity):
            print(f"Uninstalling {self.name} (version {self.installed_version})")
        subprocess.run(  # nosec
            [sys.executable, "-m", "pip", "uninstall", "-y", self.name],
            env=os.environ,
            check=False,
            capture_output=is_silent(verbosity),
        )

    def update_to_version(self, verbosity=DEFAULT_VERBOSITY) -> None:
        """
        Update command to the specified version using pip.

        If the installed version is the same as version, do nothing.

        :param verbosity: Verbosity level.
        :type verbosity: str
        """
        if not self.installed():
            self.install(verbosity=verbosity)
            return
        if self.version is None:
            # If no version is specified, we update package to its latest version
            self.update(verbosity=verbosity)
            return
        if self.installed_version_match():
            return
        # If a version is specified, we must first uninstall it
        # before installing the specified version.
        self.uninstall(verbosity=verbosity)
        self.install(verbosity=verbosity)

    def validate_contexts_match(self, *contexts: Context):
        """
        Validate that given contexts are matching command builder.

        :param contexts: Specified contexts to check matching.
        :type contexts: Context
        :raises InvalidCommand: Raised when given contexts doesn't match
            the command's requirements
        """
        missing_required_contexts = [
            required_context.name
            for required_context in self.required_contexts
            if all(
                context != required_context
                and not context.is_child_of(required_context)
                for context in contexts
            )
        ]
        if len(missing_required_contexts) != 0:
            raise InvalidCommand(
                f"Command `{self.name}` "
                "requires the following contexts, which are missing: "
                f"{', '.join(missing_required_contexts)}"
            )
        explicitly_denied_contexts = [
            context.name
            for context in contexts
            if any(
                context == denied_context or context.is_child_of(denied_context)
                for denied_context in self.denied_contexts
            )
        ]
        if len(explicitly_denied_contexts) != 0:
            raise InvalidCommand(
                f"Command `{self.name}` "
                "denies the following contexts, which are present: "
                f"{', '.join(explicitly_denied_contexts)}"
            )
        not_allowed_contexts = [
            context.name
            for context in contexts
            if not context.allowed_by_default
            and all(
                context != available_context
                and not context.is_child_of(available_context)
                for available_context in self.available_contexts
            )
        ]
        if len(not_allowed_contexts) != 0:
            raise InvalidCommand(
                f"Command `{self.name}` "
                "is not allowed due to the following contexts: "
                f"{', '.join(not_allowed_contexts)}"
            )

    def match_contexts(self, *contexts: Context) -> bool:
        """
        Check if given contexts are matching builder.

        :param contexts: Specified contexts to check matching.
        :type contexts: Context
        :return: does those contexts match builder
        :rtype: bool
        """
        try:
            self.validate_contexts_match(*contexts)
        except InvalidCommand:
            return False
        return True

    def remove_context(self, context: Context):
        """
        Remove context reference from builder.

        :param context: Context to be removed
        :type context: Context
        """
        if context in self.required_contexts:
            self.required_contexts.remove(context)
        if context in self.allowed_contexts:
            self.allowed_contexts.remove(context)
        if context in self.denied_contexts:
            self.denied_contexts.remove(context)
        if context in self.contexts_specifications:
            self.contexts_specifications.pop(context)

    def build_command(self, *contexts: Context) -> Command:
        """
        Build command according to given contexts.

        :param contexts: Specified contexts to build command according to.
        :type contexts: Context
        :return: Built command.
        :rtype: Command
        """
        self.validate_contexts_match(*contexts)
        return Command(name=self.name, args=self.build_args(*contexts))

    def build_args(self, *contexts: Context) -> List[str]:
        """
        Build arguments for command according to contexts.

        :param contexts: Specified contexts to arguments
            command according to.
        :type contexts: Context
        :return: Arguments list
        :rtype: List[str]
        """
        args = list(self.default_args)
        for context in contexts:
            args = self.get_context_specification(context).update_args(args)
        return args

    def get_context_specification(self, context: Context) -> ContextSpecification:
        """
        Get context specification from context name.

        :param context: Specified context
        :type context: Context
        :return: Context specification to build command according to
        :rtype: ContextSpecification
        """
        return self.contexts_specifications.get(context, ContextSpecification())

    def as_dict(self) -> OrderedDictType[str, Any]:
        """
        Encode command builder as a dictionary.

        This is used in order to serialize the command builder in
        a configuration file.

        :return: Serialized representation dictionary
        :rtype: OrderedDict[str, Any]
        """
        builder_as_dict: OrderedDict[str, Any] = OrderedDict()
        builder_as_dict[HELP] = self.help
        if len(self.default_args) != 0:
            builder_as_dict[ARGS] = self.default_args
        if len(self.required_contexts) != 0:
            required_contexts = [context.name for context in self.required_contexts]
            required_contexts.sort()
            builder_as_dict[REQUIRED_CONTEXTS] = required_contexts
        if len(self.allowed_contexts) != 0:
            allowed_contexts = [context.name for context in self.allowed_contexts]
            allowed_contexts.sort()
            builder_as_dict[ALLOWED_CONTEXTS] = allowed_contexts
        if len(self.denied_contexts) != 0:
            denied_contexts = [context.name for context in self.denied_contexts]
            denied_contexts.sort()
            builder_as_dict[DENIED_CONTEXTS] = denied_contexts
        if self.version is not None:
            builder_as_dict[VERSION] = self.version
        specified_contexts = list(self.specified_contexts)
        specified_contexts.sort(key=lambda context: context.name)
        for context in specified_contexts:
            builder_as_dict[context.name] = self.contexts_specifications[
                context
            ].as_dict()
        return builder_as_dict

    @classmethod
    def from_dict(
        cls,
        command_name: str,
        builder_setups: Dict[str, Any],
        contexts_repository: ContextsRepository,
    ):
        """
        Build command builder according to a given configuration.

        :param command_name: Name of the command to be built
        :type command_name: str
        :param builder_setups: Command builder configuration
        :type builder_setups: Dict[str, Any]
        :param contexts_repository: contexts repository to get contexts from
        :type contexts_repository: ContextsRepository
        :return: Command builder as specified
        :rtype: CommandBuilder
        :raises MissingHelpString: Raised when help string is missing
        """
        if HELP not in builder_setups:
            raise MissingHelpString(location=[command_name])
        return CommandBuilder(
            name=command_name,
            help=builder_setups[HELP],
            default_args=builder_setups.get(ARGS, []),
            version=builder_setups.get(VERSION),
            required_contexts=cls.build_contexts_list(
                command_name=command_name,
                key_name=REQUIRED_CONTEXTS,
                builder_setups=builder_setups,
                contexts_repository=contexts_repository,
            ),
            allowed_contexts=cls.build_contexts_list(
                command_name=command_name,
                key_name=ALLOWED_CONTEXTS,
                builder_setups=builder_setups,
                contexts_repository=contexts_repository,
            ),
            denied_contexts=cls.build_contexts_list(
                command_name=command_name,
                key_name=DENIED_CONTEXTS,
                builder_setups=builder_setups,
                contexts_repository=contexts_repository,
            ),
            contexts_specifications=cls.build_contexts_specifications(
                command_name=command_name,
                builder_setups=builder_setups,
                contexts_repository=contexts_repository,
            ),
        )

    @classmethod
    def build_contexts_list(
        cls,
        command_name: str,
        key_name: str,
        builder_setups: Dict[str, Any],
        contexts_repository: ContextsRepository,
    ) -> List[Context]:
        """
        Read contexts list from dictionary.

        :param command_name: Name of the command to be built
        :type command_name: str
        :param key_name: Key of the desired contexts list
        :type key_name: str
        :param builder_setups: Command builder configuration
        :type builder_setups: Dict[str, Any]
        :param contexts_repository: contexts repository to get contexts from
        :type contexts_repository: ContextsRepository
        :return: Contexts list
        :rtype: List[Context]
        :raises InvalidConfiguration: Raised when one of the contexts was not
            set in the contexts repository
        """
        if key_name not in builder_setups:
            return []
        context_names = builder_setups[key_name]
        try:
            return [contexts_repository[context_name] for context_name in context_names]
        except UnknownContext as error:
            raise InvalidConfiguration(
                message="Unknown context in configuration",
                location=[command_name, key_name, error.context_name],
            ) from error

    @classmethod
    def build_contexts_specifications(
        cls,
        command_name: str,
        builder_setups: Dict[str, Any],
        contexts_repository: ContextsRepository,
    ) -> Dict[Context, ContextSpecification]:
        """
        Read command builder's contexts specifications dictionary from configuration.

        :param command_name: Name of the command to be built
        :type command_name: str
        :param builder_setups: Command builder configuration
        :type builder_setups: Dict[str, Any]
        :param contexts_repository: contexts repository to get contexts from
        :type contexts_repository: ContextsRepository
        :return: Contexts specifications dictionary
        :rtype: Dict[Context, ContextSpecification]
        :raises InvalidConfiguration: Raised when one of the contexts was not
            set in the contexts repository
        :raises StatueConfigurationError: Raised when there is an inconsistency
            in a context specification definition.
        """
        context_names = [
            context_name
            for context_name in builder_setups.keys()
            if context_name not in cls.setup_words()
        ]
        contexts_specifications = {}
        for context_name in context_names:
            try:
                contexts_specifications[
                    contexts_repository[context_name]
                ] = ContextSpecification.from_dict(
                    context_specification_setups=builder_setups[context_name]
                )
            except UnknownContext as error:
                raise InvalidConfiguration(
                    message="Unknown context in configuration",
                    location=[command_name, error.context_name],
                ) from error
            except StatueConfigurationError as error:
                error.append_location_item(context_name)
                error.append_location_item(command_name)
                raise error
        return contexts_specifications

    @classmethod
    def setup_words(cls) -> List[str]:
        """
        Predefined setup words.

        :return: List of predefined setup words
        :rtype: List[str]
        """
        return [
            HELP,
            ARGS,
            VERSION,
            ALLOWED_CONTEXTS,
            DENIED_CONTEXTS,
            REQUIRED_CONTEXTS,
        ]

    def _validate_consistency(self, **kwargs: Set[Context]):
        assert set(kwargs.keys()) == {  # nosec
            "allowed",
            "denied",
            "required",
            "specified",
        }
        for key1, key2 in itertools.combinations(kwargs.keys(), 2):
            value1, value2 = kwargs[key1], kwargs[key2]
            values_intersection = value1.intersection(value2)
            if len(values_intersection) != 0:
                self._raise_inconsistency_error(
                    type1=key1,
                    type2=key2,
                    contexts=values_intersection,
                )

    def _raise_inconsistency_error(
        self, type1: str, type2: str, contexts: Set[Context]
    ):
        context_names = [context.name for context in contexts]
        context_names.sort()
        message = f"{type1} and {type2} contexts clash"
        location = [self.name, f"{type1}/{type2}"]
        if len(context_names) == 1:
            location.append(context_names[0])
        raise InconsistentConfiguration(message, location=location)

    def _initialize_contexts(self):
        self.required_contexts = set()
        self.allowed_contexts = set()
        self.denied_contexts = set()
        self.contexts_specifications = {}

    def _get_package(self):  # pragma: no cover
        """
        Get package of the desired command.

        If package is not installed, returns None.

        :return: self package
        """
        importlib.reload(pkg_resources)
        for package in list(pkg_resources.working_set):
            if package.key == self.name:
                return package
        return None



"""Utility methods for CLI."""
from typing import List, Optional, Sequence, TypeVar

T = TypeVar("T")


def list_or_none(some_list: Optional[Sequence[T]]) -> Optional[List[T]]:
    """
    Return None if sequence is empty, else return sequence as list.

    :param some_list: Optional sequence
    :type some_list: Optional[Sequence[Any]]
    :return: None if empty, list of not
    :rtype: Optional[List[T]]
    """
    if some_list is None or len(some_list) == 0:
        return None
    return list(some_list)



"""Command map runner."""
import abc
import asyncio
import time
from enum import Enum, auto
from pathlib import Path
from typing import List

import tqdm

from statue.command import Command
from statue.commands_map import CommandsMap
from statue.constants import BAR_FORMAT, MAIN_BAR_COLOR, SECONDARY_BAR_COLOR
from statue.evaluation import Evaluation, SourceEvaluation


class RunnerMode(Enum):
    """Enum indicating in which mode are we running evaluation."""

    SYNC = auto()
    ASYNC = auto()
    DEFAULT_MODE = SYNC


class EvaluationRunner:  # pylint: disable=too-few-public-methods
    """Evaluation runner interface."""

    @abc.abstractmethod
    def evaluate(
        self,
        commands_map: CommandsMap,
    ) -> Evaluation:
        """
        Abstract evaluation method.

        # noqa: DAR202

        :param commands_map: map from source file to list of commands to run on it
        :type commands_map: CommandsMap
        :return: Total evaluation after running all commands.
        :rtype: Evaluation
        """


class SynchronousEvaluationRunner(  # pylint: disable=too-few-public-methods
    EvaluationRunner
):
    """Runner class for running commands synchronously."""

    def evaluate(
        self,
        commands_map: CommandsMap,
    ) -> Evaluation:
        """
        Run commands map and return evaluation report.

        :param commands_map: map from source file to list of commands to run on it
        :type commands_map: CommandsMap
        :return: Total evaluation after running all commands.
        :rtype: Evaluation
        """
        evaluation = Evaluation()
        total_start_time = time.time()
        with tqdm.trange(
            commands_map.total_commands_count,
            bar_format=BAR_FORMAT,
            colour=MAIN_BAR_COLOR,
        ) as main_bar:
            for source, commands in commands_map.items():
                source_start_time = time.time()
                evaluation[source] = SourceEvaluation()
                for command in tqdm.tqdm(
                    commands,
                    bar_format=BAR_FORMAT,
                    colour=SECONDARY_BAR_COLOR,
                    leave=False,
                    desc=str(source),
                ):
                    evaluation[source].append(command.execute(source))
                    main_bar.update(1)
                source_end_time = time.time()
                evaluation[source].source_execution_duration = (
                    source_end_time - source_start_time
                )
        total_end_time = time.time()
        evaluation.total_execution_duration = total_end_time - total_start_time
        return evaluation


class AsynchronousEvaluationRunner(EvaluationRunner):
    """Runner class for running commands asynchronously."""

    def __init__(self):
        """Initialize runner."""
        self.update_lock = asyncio.Lock()

    def evaluate(
        self,
        commands_map: CommandsMap,
    ) -> Evaluation:
        """
        Run commands map asynchronously and return evaluation report.

        :param commands_map: map from source file to list of commands to run on it
        :type commands_map: CommandsMap
        :return: Total evaluation after running all commands.
        :rtype: Evaluation
        """
        return asyncio.run(self.evaluate_commands_map(commands_map))

    async def evaluate_commands_map(
        self,
        commands_map: CommandsMap,
    ):
        """
        Main async function to run commands map and return evaluation report.

        :param commands_map: map from source file to list of commands to run on it
        :type commands_map: CommandsMap
        :return: Evaluation
        """
        evaluation = Evaluation()
        start_time = time.time()
        max_source_name_length = max(
            len(source.as_posix()) for source in commands_map.keys()
        )
        with tqdm.trange(
            commands_map.total_commands_count,
            bar_format=BAR_FORMAT,
            colour=MAIN_BAR_COLOR,
        ) as main_bar:
            coros = [
                self.evaluate_source(
                    source_bar_pos=pos,
                    source=source,
                    commands=commands,
                    evaluation=evaluation,
                    main_bar=main_bar,
                    max_source_name_length=max_source_name_length,
                )
                for pos, (source, commands) in enumerate(commands_map.items(), start=1)
            ]
            await asyncio.gather(*coros)
        end_time = time.time()
        evaluation.total_execution_duration = end_time - start_time
        return evaluation

    async def evaluate_source(  # pylint: disable=too-many-arguments
        self,
        source: Path,
        commands: List[Command],
        evaluation: Evaluation,
        main_bar: tqdm.tqdm,
        source_bar_pos: int,
        max_source_name_length: int,
    ):
        """
        Evaluate commands on source and return source evaluation report.

        :param source_bar_pos: Position of the source bar to print
        :type source_bar_pos: int
        :param source: Path of the desired source.
        :type source: Path
        :param commands: List of commands to run on the source.
        :type commands: List[Command]
        :param evaluation: Evaluation instance to be updated after commands are running.
        :type evaluation: Evaluation
        :param main_bar: progress bar that shows how far are we in evaluating the source
        :type main_bar: tqdm.tqdm
        :param max_source_name_length: Maximum source name length
        :type max_source_name_length: int
        """
        evaluation[source] = SourceEvaluation()
        start_time = time.time()
        with tqdm.trange(
            len(commands),
            bar_format=BAR_FORMAT,
            position=source_bar_pos,
            leave=False,
            colour=SECONDARY_BAR_COLOR,
            desc=f"{source.as_posix():{max_source_name_length}}",
        ) as source_bar:
            coros = [
                self.evaluate_command(
                    command=command,
                    source=source,
                    evaluation=evaluation,
                    source_bar=source_bar,
                    main_bar=main_bar,
                )
                for command in commands
            ]
            await asyncio.gather(*coros)
        end_time = time.time()
        evaluation[source].source_execution_duration = end_time - start_time
        await self.update_lock.acquire()
        self.update_lock.release()

    async def evaluate_command(  # pylint: disable=too-many-arguments
        self,
        command: Command,
        source: Path,
        evaluation: Evaluation,
        source_bar: tqdm.tqdm,
        main_bar: tqdm.tqdm,
    ):
        """
        Evaluate command on source and return command evaluation report.

        :param source: Path of the desired source.
        :type source: Path
        :param command: Command to run on the source.
        :type command: Command
        :param evaluation: Evaluation instance to be updated after commands are running.
        :type evaluation: Evaluation
        :param source_bar: tqdm progress bar to show the progress
            of evaluating this specific source.
        :type source_bar: tqdm.tqdm
        :param main_bar: tqdm progress bar to show total progress
        :type main_bar: tqdm.tqdm
        """
        command_evaluation = await command.execute_async(source)
        await self.update_lock.acquire()
        evaluation[source].append(command_evaluation)
        source_bar.update(1)
        main_bar.update(1)
        self.update_lock.release()


MODE_TO_RUNNER_DICT = {
    RunnerMode.SYNC.name: SynchronousEvaluationRunner,
    RunnerMode.ASYNC.name: AsynchronousEvaluationRunner,
}


def build_runner(runner_mode: str) -> EvaluationRunner:
    """
    Build commands runner.

    :param runner_mode: Which mode should the runner work in
    :type runner_mode: str
    :return: Runner instance.
    :rtype: EvaluationRunner
    """
    return MODE_TO_RUNNER_DICT[runner_mode]()



"""Show configuration CLI."""
from pathlib import Path
from typing import Optional

import click

from statue.cli.common_flags import config_path_option
from statue.cli.config.config_cli import config_cli
from statue.config.configuration import Configuration
from statue.constants import ENCODING


@config_cli.command("show")
@config_path_option
def show_config_cli(config: Optional[Path]):
    """Show configuration file context."""
    if config is None:
        config = Configuration.configuration_path()
    with open(config, mode="r", encoding=ENCODING) as config_file:
        lines = config_file.readlines()
    click.echo_via_pager(lines)



"""Place for saving all available command builders."""
from collections import OrderedDict
from typing import Any, Dict, Iterator, List
from typing import OrderedDict as OrderedDictType

from statue.command_builder import CommandBuilder
from statue.config.contexts_repository import ContextsRepository
from statue.exceptions import UnknownCommand


class CommandsRepository:
    """Repository class for saving and accessing command builders."""

    def __init__(self, *command_builders: CommandBuilder):
        """
        Constructor.

        :param command_builders: Initial command builders to be saved
        :type command_builders: Context
        """
        self.command_builders_map: Dict[str, CommandBuilder] = {}
        self.add_command_builders(*command_builders)

    def __len__(self) -> int:
        """
        Number of available commands.

        :return: Number of commands
        :rtype: int
        """
        return len(self.command_builders_map)

    def __iter__(self) -> Iterator[CommandBuilder]:
        """
        Iterate over all available command builders.

        :return: Command builder iterator
        :rtype: Iterator[CommandBuilder]
        """
        return iter(self.command_builders_map.values())

    def __getitem__(self, item: str) -> CommandBuilder:
        """
        Get command builder by name or alias.

        :param item: Command name to be retrieved
        :type item: str
        :return: Command builder with given name
        :rtype: CommandBuilder
        :raises UnknownCommand: Raised when command is not found
        """
        if item in self:
            return self.command_builders_map[item]
        raise UnknownCommand(command_name=item)

    def __contains__(self, item) -> bool:
        """
        Does command builder available in repository.

        :param item: Command name to be searched
        :type item: str
        :return: does the command exist in repository
        :rtype: bool
        """
        return item in self.command_builders_map

    @property
    def command_names_list(self) -> List[str]:
        """Names of all available commands."""
        return list(self.command_builders_map.keys())

    def add_command_builders(self, *command_builders: CommandBuilder):
        """
        Add command builders to repository.

        :param command_builders: Command builder to be added to the
            repository
        :type command_builders: CommandBuilder
        """
        for command_builder in command_builders:
            self.command_builders_map[command_builder.name] = command_builder

    def remove_command_builder(self, command_builder: CommandBuilder):
        """
        Remove a command builder from repository.

        :param command_builder: Command builder to be removed.
        :type command_builder: CommandBuilder
        """
        del self.command_builders_map[command_builder.name]

    def reset(self):
        """Clear repository from all command builders."""
        self.command_builders_map.clear()

    def as_dict(self) -> OrderedDictType[str, Any]:
        """
        Encode commands repository as a dictionary.

        This is used in order to serialize the commands repository in
        a configuration file.

        :return: Serialized representation dictionary
        :rtype: OrderedDict[str, Any]
        """
        command_builders_list = list(self)
        command_builders_list.sort(key=lambda commands_builder: commands_builder.name)
        return OrderedDict(
            [
                (command_builder.name, command_builder.as_dict())
                for command_builder in command_builders_list
            ]
        )

    @classmethod
    def from_dict(
        cls, config: Dict[str, Any], contexts_repository: ContextsRepository
    ) -> "CommandsRepository":
        """
        Create commands repository from given configuration.

        :param config: Configuration to update repository from
        :type config: Dict[str, Any]
        :param contexts_repository: Contexts repository to get contexts from
        :type contexts_repository: ContextsRepository
        :return: Commands repository as described in configuration
        :rtype: CommandsRepository
        """
        builders = [
            CommandBuilder.from_dict(
                command_name=command_name,
                builder_setups=builder_setups,
                contexts_repository=contexts_repository,
            )
            for command_name, builder_setups in config.items()
        ]
        return CommandsRepository(*builders)



"""Constants module."""

STATUE = "STATUE"
ENCODING = "utf-8"

DEFAULT_HISTORY_SIZE = 30

GENERAL = "general"
COMMANDS = "commands"
CONTEXTS = "contexts"
SOURCES = "sources"

HELP = "help"
ARGS = "args"
CLEAR_ARGS = "clear_args"
ADD_ARGS = "add_args"
ALLOW_LIST = "allow_list"
DENY_LIST = "deny_list"
ALLOWED_CONTEXTS = "allowed_contexts"
DENIED_CONTEXTS = "denied_contexts"
REQUIRED_CONTEXTS = "required_contexts"
ALIASES = "aliases"
PARENT = "parent"
ALLOWED_BY_DEFAULT = "allowed_by_default"
VERSION = "version"
MODE = "mode"
HISTORY_SIZE = "history_size"
CACHE = "cache"

TEMPLATE_NAME_REGEX = r"^[A-Za-z]\w*$"
DATETIME_FORMAT = "%m/%d/%Y, %H:%M:%S"
BAR_FORMAT = "{l_bar}{bar}| {n_fmt}/{total_fmt}"
MAIN_BAR_COLOR = "blue"
SECONDARY_BAR_COLOR = "yellow"



"""Templates CLI."""
import sys

import click

from statue.cli.cli import statue_cli
from statue.cli.common_flags import config_path_option
from statue.cli.styled_strings import bullet_style, failure_style, name_style
from statue.config.configuration import Configuration
from statue.exceptions import StatueTemplateError, UnknownTemplate
from statue.templates.templates_provider import TemplatesProvider


@statue_cli.group("templates")
def templates_cli() -> None:
    """Template related actions such as list, show, etc."""


@templates_cli.command("list")
def list_templates_cli():
    """List all available templates."""
    click.echo(bullet_style("Default templates:"))
    for template_name in TemplatesProvider.default_templates_names():
        click.echo(f"\t{name_style(template_name)}")
    user_templates = TemplatesProvider.user_templates_names()
    if len(user_templates) == 0:
        click.echo("No user templates.")
        return
    click.echo(bullet_style("User templates:"))
    for template_name in user_templates:
        click.echo(f"\t{name_style(template_name)}")


@templates_cli.command("show")
@click.argument("template_name", type=str)
def show_templates_cli(template_name):
    """Show template by name."""
    try:
        template_path = TemplatesProvider.get_template_path(template_name)
    except UnknownTemplate as error:
        click.echo(failure_style(str(error)))
        sys.exit(1)
    with template_path.open(mode="r") as template_file:
        template_lines = template_file.readlines()
    click.echo("".join(template_lines))


@templates_cli.command("save")
@click.argument("template_name", type=str)
@config_path_option
@click.option(
    "--override/-no-override",
    is_flag=True,
    default=False,
    help="Override template with given name if already exists.",
)
def save_templates_cli(template_name, config, override):
    """Save current configuration as template."""
    if config is None:
        config = Configuration.configuration_path()
    configuration = Configuration.from_file(config)
    configuration.sources_repository.reset()
    try:
        TemplatesProvider.save_template(
            name=template_name, configuration=configuration, override=override
        )
    except StatueTemplateError as error:
        click.echo(failure_style(str(error)))
        sys.exit(1)
    click.echo(f'Template "{template_name}" was saved successfully!')


@templates_cli.command("remove")
@click.argument("template_name", type=str)
def remove_templates_cli(template_name):
    """Remove template by name."""
    if template_name in TemplatesProvider.default_templates_names():
        click.echo(failure_style("Cannot remove a default template."))
        sys.exit(1)
    if template_name not in TemplatesProvider.user_templates_names():
        click.echo(failure_style(f'Could not find template named "{template_name}"'))
        sys.exit(1)
    if not click.confirm(
        f'Are you sure you want to remove template "{template_name}"?', default=False
    ):
        click.echo(failure_style("Abort!"))
        sys.exit(1)
    TemplatesProvider.remove_template(template_name)
    click.echo(f'Template "{template_name}" was saved successfully!')


@templates_cli.command("clear")
def clear_user_templates_cli():
    """Remove template by name."""
    templates_number = len(TemplatesProvider.user_templates_names())
    if templates_number == 0:
        click.echo("No templates to remove.")
        return
    if not click.confirm(
        f"Are you sure you want to remove {templates_number} templates?", default=False
    ):
        click.echo(failure_style("Abort!"))
        sys.exit(1)
    TemplatesProvider.clear_user_templates()
    click.echo("All user templates were successfully remove!")



"""Verbosity related constants and methods."""
SILENT = "silent"
VERBOSE = "verbose"
NORMAL = "normal"

VERBOSITIES = [NORMAL, SILENT, VERBOSE]
DEFAULT_VERBOSITY = NORMAL


def is_silent(verbosity: str) -> bool:
    """
    Validates if verbosity is silent.

    :param verbosity: String. verbosity value
    :return: Boolean. is verbosity silent
    """
    return verbosity == SILENT


def is_verbose(verbosity: str) -> bool:
    """
    Validates if verbosity is verbose.

    :param verbosity: String. verbosity value
    :return: Boolean. is verbosity verbose
    """
    return verbosity == VERBOSE



"""Add contexts to configuration interactively."""
from typing import List, Optional

import click
import click_params as clickp

from statue.cli.config.interactive_adders.adders_utils import get_help_string
from statue.cli.styled_strings import failure_style, name_style
from statue.config.contexts_repository import ContextsRepository
from statue.context import Context


class InteractiveContextAdder:
    """Singleton class for adding contexts to configuration instance."""

    @classmethod
    def add_context(cls, contexts_repository: ContextsRepository):
        """
        Add context interactively to context repository.

        :param contexts_repository: contexts repository to add context to
        :type contexts_repository: ContextsRepository
        """
        name = cls.get_context_name(contexts_repository)
        help_string = get_help_string(name)
        aliases = cls.get_aliases(name=name, contexts_repository=contexts_repository)
        parent = cls.get_parent(
            name=name, aliases=aliases, contexts_repository=contexts_repository
        )
        allowed_by_default = click.confirm(
            f"Would you like {name_style(name)} context to be allowed by default?",
            default=False,
            show_default=True,
        )
        context = Context(
            name=name,
            help=help_string,
            aliases=aliases,
            parent=parent,
            allowed_by_default=allowed_by_default,
        )
        contexts_repository.add_contexts(context)

    @classmethod
    def edit_context(cls, name: str, contexts_repository: ContextsRepository):
        """
        Edit existing context in context repository.

        :param name: Name of the context to edit
        :type name: str
        :param contexts_repository: Contexts repository instance to edit context in
        :type contexts_repository: ContextsRepository
        """
        context = contexts_repository[name]
        context.clear_aliases()
        context.help = get_help_string(name)
        context.aliases = cls.get_aliases(
            name=name, contexts_repository=contexts_repository
        )
        context.parent = cls.get_parent(
            name=name, aliases=context.aliases, contexts_repository=contexts_repository
        )
        context.allowed_by_default = click.confirm(
            f"Would you like {name_style(name)} context to be allowed by default?",
            default=False,
            show_default=True,
        )

    @classmethod
    def get_context_name(cls, contexts_repository: ContextsRepository) -> str:
        """
        Get context name interactively from user.

        :param contexts_repository: Context repository to check pre-existing contexts in
        :type contexts_repository: ContextsRepository
        :return: new context name
        :rtype: str
        """
        name = ""
        while name == "":
            name = click.prompt(
                "Please choose context name",
                default="",
                show_default=False,
            )
            name = name.strip()
            if name in contexts_repository:
                click.echo(failure_style(f"{name} already exists in repository!"))
                name = ""
        return name

    @classmethod
    def get_aliases(
        cls, name: str, contexts_repository: ContextsRepository
    ) -> List[str]:
        """
        Get aliases for context.

        :param name: name of the context to get aliases for
        :type name: str
        :param contexts_repository: Context repository to check pre-existing contexts in
        :type contexts_repository: ContextsRepository
        :return: new context aliases
        :rtype: List[str]
        """
        aliases = None
        while aliases is None:
            aliases = click.prompt(
                f"Please add aliases of {name_style(name)} separated by commas "
                "(press enter to skip)",
                default="",
                type=clickp.StringListParamType(),
                show_default=False,
            )
            aliases = [alias.strip() for alias in aliases if alias.strip() != ""]
            if name in aliases:
                aliases.remove(name)
            predefined_aliases = [
                alias for alias in aliases if alias in contexts_repository
            ]
            if len(predefined_aliases) != 0:
                click.echo(
                    failure_style(
                        "Could not set the following aliases since they are "
                        f"already taken: {','.join(predefined_aliases)}"
                    )
                )
                aliases = None
        return aliases

    @classmethod
    def get_parent(
        cls,
        name: str,
        aliases: List[str],
        contexts_repository: ContextsRepository,
    ) -> Optional[Context]:
        """
        Get parent for the new context.

        :param name: Context name to get parent for
        :type name: str
        :param aliases: Aliases list for the context
        :type aliases: List[str]
        :param contexts_repository: Contexts repository to check existing contexts
        :type contexts_repository: ContextsRepository
        :return: Optional context parent
        :rtype: Optional[Context]
        """
        parent_name = None
        while parent_name is None:
            parent_name = click.prompt(
                "Please choose parent context (press enter to skip)",
                type=str,
                default="",
                show_default=False,
            )
            parent_name = parent_name.strip()
            if parent_name == "":
                return None
            if parent_name == name or parent_name in aliases:
                click.echo(failure_style(f"{name} can't be parent of itself"))
                parent_name = None
                continue
            if parent_name not in contexts_repository:
                click.echo(failure_style(f"{parent_name} doesn't exist."))
                parent_name = None
                continue
        return contexts_repository[parent_name]



"""Show commands map as tree."""
from typing import List, Optional, Sequence

import click

from statue.cli.cli import pass_configuration, statue_cli
from statue.cli.cli_util import list_or_none
from statue.cli.common_flags import allow_option, contexts_option, deny_option
from statue.cli.styled_strings import bullet_style, name_style, source_style
from statue.commands_filter import CommandsFilter
from statue.config.configuration import Configuration


@statue_cli.command("show-tree", short_help="Show sources configuration as a tree.")
@pass_configuration
@contexts_option
@allow_option
@deny_option
def show_tree_cli(
    configuration: Configuration,
    context: Sequence[str],
    allow: Optional[List[str]],
    deny: Optional[List[str]],
):
    """
    Show sources configuration as a tree.

    This method prints the sources' configuration as a tree, including:
    contexts, allow and deny lists and matching commands.
    """
    sources_list = configuration.sources_repository.sources_list
    general_filter = CommandsFilter(
        allowed_commands=list_or_none(allow),
        denied_commands=list_or_none(deny),
        contexts=[
            configuration.contexts_repository[context_name] for context_name in context
        ],
    )
    if len(sources_list) == 0:
        click.echo("No sources configuration is specified.")
    for source in sources_list:
        source_commands_filter = CommandsFilter.merge(
            general_filter, configuration.sources_repository[source]
        )
        context_names = [context.name for context in source_commands_filter.contexts]
        click.echo(
            f"{source_style(str(source))} "
            f"({bullet_style('contexts')}: "
            f"{__join_names(context_names)}, "
            f"{bullet_style('allowed')}: "
            f"{__join_names(source_commands_filter.allowed_commands)}, "
            f"{bullet_style('denied')}: "
            f"{__join_names(source_commands_filter.denied_commands)}):"
        )
        commands = configuration.build_commands(source_commands_filter)
        click.echo(f"\t{__join_names([command.name for command in commands])}")


def __join_names(names_list):
    if names_list is None or len(names_list) == 0:
        return "empty"
    names_list = list(names_list)
    names_list.sort()
    return ", ".join([name_style(name) for name in names_list])



"""Configuration templates provider."""
import re
import sys
from pathlib import Path
from typing import Dict, Set, Union

from statue.config.configuration import Configuration
from statue.constants import TEMPLATE_NAME_REGEX
from statue.exceptions import StatueTemplateError, UnknownTemplate

if sys.version_info < (3, 9):  # pragma: no cover
    import importlib_resources as resources
    from importlib_resources.abc import Traversable
else:  # pragma: no cover
    from importlib import resources
    from importlib.abc import Traversable


TemplateSourceType = Union[Traversable, Path]


class TemplatesProvider:
    """A singleton for retrieving configuration templates."""

    @classmethod
    def default_templates_map(cls) -> Dict[str, Traversable]:
        """
        Creates map from default template name to path.

        :return: templates map.
        :rtype: Dict[str, Traversable]
        """
        return {
            cls.get_template_name(path): path
            for path in TemplatesProvider.default_templates()
        }

    @classmethod
    def user_templates_map(cls) -> Dict[str, Path]:
        """
        Creates map from user template name to path.

        :return: templates map.
        :rtype: Dict[str, Path]
        """
        return {
            cls.get_template_name(path): path
            for path in TemplatesProvider.user_templates()
        }

    @classmethod
    def templates_map(cls) -> Dict[str, TemplateSourceType]:
        """
        Creates map from template name to path.

        :return: templates map.
        :rtype: Dict[str, Traversable]
        """
        return {**cls.default_templates_map(), **cls.user_templates_map()}

    @classmethod
    def default_templates(cls) -> Set[Traversable]:
        """
        Get all default templates available.

        :return: default templates
        :rtype: Set[Traversable]
        """
        return {
            path
            for path in resources.files("statue.templates.default_templates").iterdir()
            if path.name.split(".")[-1] == "toml"
        }

    @classmethod
    def user_templates(cls) -> Set[Path]:
        """
        Get all user templates available.

        :return: User templates
        :rtype: Set[Path]
        """
        return set(cls.user_templates_directory().iterdir())

    @classmethod
    def default_templates_names(cls) -> Set[str]:
        """
        Get all default template names.

        :return: all default template names
        :rtype: Set[str]
        """
        return {cls.get_template_name(path) for path in cls.default_templates()}

    @classmethod
    def user_templates_names(cls) -> Set[str]:
        """
        Get all user template names.

        :return: all user template names
        :rtype: Set[str]
        """
        return {cls.get_template_name(path) for path in cls.user_templates()}

    @classmethod
    def all_template_names(cls) -> Set[str]:
        """
        Get all available template names.

        :return: all template names
        :rtype: Set[str]
        """
        return {*cls.default_templates_names(), *cls.user_templates_names()}

    @classmethod
    def get_template_path(cls, template_name: str) -> Traversable:
        """
        Get a template path by name.

        :param template_name: name of the template path to retrieve
        :type template_name: str
        :return: requested template path
        :rtype: Traversable
        :raises UnknownTemplate: raised when given template name is unknown
        """
        try:
            return cls.templates_map()[template_name]
        except KeyError as error:
            raise UnknownTemplate(template_name) from error

    @classmethod
    def user_templates_directory(cls) -> Path:
        """
        Get user templates directory.

        :return: User templates directory
        :rtype: Path
        """
        templates_directory = Path.home() / ".statue" / "templates"
        templates_directory.mkdir(parents=True, exist_ok=True)
        return templates_directory

    @classmethod
    def get_template_name(cls, path: TemplateSourceType) -> str:
        """
        Get template name from path.

        :param path: Path of the template
        :type path: TemplateSourceType
        :return: Template name
        :rtype: str
        """
        return path.name.split(".")[0]

    @classmethod
    def save_template(
        cls, name: str, configuration: Configuration, override: bool = False
    ):
        """
        Save configuration as template with given name.

        :param name: Name of the template to save
        :type name: str
        :param configuration: Configuration to be saved as template
        :type configuration: Configuration
        :param override: Should override existing template. False by default.
        :type override: bool
        :raises StatueTemplateError: raised when configuration cannot be
            saved as a template.
        """
        if not re.match(TEMPLATE_NAME_REGEX, name):
            raise StatueTemplateError(
                f'"{name}" is an invalid template name. '
                "Templates should start with a letter and contain only "
                "letters, numbers and underscores"
            )
        if not override and name in cls.all_template_names():
            raise StatueTemplateError(f'"{name}" template is already taken.')
        configuration.to_toml(cls.user_templates_directory() / f"{name}.toml")

    @classmethod
    def remove_template(cls, name: str):
        """
        Remove user defined template.

        :param name: Name of the template to be removed
        :type name: str
        :raises StatueTemplateError: Raised when trying to remove a default template
        :raises UnknownTemplate: Raised when there is no template with given name
        """
        if name in cls.default_templates_names():
            raise StatueTemplateError("Default templates cannot be removed.")
        try:
            template_path = cls.user_templates_map()[name]
            template_path.unlink()
        except KeyError as error:
            raise UnknownTemplate(name) from error

    @classmethod
    def clear_user_templates(cls):
        """Remove all user templates."""
        for template_path in cls.user_templates():
            template_path.unlink()



"""Commands configuration CLI."""
import sys
from pathlib import Path
from typing import Optional

import click

from statue.cli.common_flags import config_path_option, verbose_option
from statue.cli.config.config_cli import config_cli
from statue.cli.config.interactive_adders.interactive_command_adder import (
    InteractiveCommandAdder,
)
from statue.cli.styled_strings import failure_style, name_style
from statue.config.configuration import Configuration
from statue.exceptions import UnknownCommand


@config_cli.command("add-command")
@config_path_option
def add_command_cli(config):
    """Add new command to configuration."""
    if config is None:
        config = Configuration.configuration_path()
    configuration = Configuration.from_file(config)
    InteractiveCommandAdder.add_command(configuration)
    configuration.to_toml(config)
    click.echo("Command was successfully added!")


@config_cli.command("edit-command")
@config_path_option
@click.argument("command_name", type=str)
def edit_command_cli(config, command_name):
    """Edit context from configuration."""
    if config is None:
        config = Configuration.configuration_path()
    configuration = Configuration.from_file(config)
    InteractiveCommandAdder.edit_command(name=command_name, configuration=configuration)
    configuration.to_toml(config)
    click.echo(f"{name_style(command_name)} was successfully edited!")


@config_cli.command("remove-command")
@config_path_option
@click.argument("command_name", type=str)
def remove_command_cli(config, command_name):
    """Remove command from configuration."""
    if config is None:
        config = Configuration.configuration_path()
    configuration = Configuration.from_file(config)
    try:
        command_builder = configuration.commands_repository[command_name]
    except UnknownCommand as error:
        click.echo(failure_style(str(error)))
        sys.exit(1)
    if not click.confirm(
        f"Are you sure you would like to remove the command {name_style(command_name)} "
        "and all of its references from configuration?",
    ):
        click.echo("Abort!")
        return
    configuration.remove_command(command_builder)
    configuration.to_toml(config)
    click.echo(f"{name_style(command_name)} was successfully removed!")


@config_cli.command(
    "fix-versions", short_help="Fixate the installed version of the commands."
)
@config_path_option
@click.option(
    "-l",
    "--latest",
    is_flag=True,
    default=False,
    help=(
        "Update commands to latest before fixing the version. "
        "If a command is not installed, will install it."
    ),
)
@verbose_option
def fixate_commands_versions_cli(
    config: Optional[Path],
    latest: bool,
    verbosity: str,
):
    """
    Fixate the installed version of the commands.

    This helps you make sure that you use the same checkers in all commands
    across time.
    """
    if config is None:
        config = Configuration.configuration_path()
    configuration = Configuration.from_file(config)
    if len(configuration.commands_repository) == 0:
        click.echo("No commands to fixate.")
        return
    for command_builder in configuration.commands_repository:
        if latest:
            command_builder.update(verbosity=verbosity)
        if not command_builder.installed():
            continue
        command_builder.set_version_as_installed()
    configuration.to_toml(config)



"""CLI definitions."""
from statue.cli.cli import statue_cli
from statue.cli.commands import commands_cli
from statue.cli.config import config_cli
from statue.cli.contexts import context_cli
from statue.cli.history import history_cli
from statue.cli.run import run_cli
from statue.cli.templates import templates_cli

__all__ = [
    "statue_cli",
    "commands_cli",
    "config_cli",
    "context_cli",
    "run_cli",
    "history_cli",
    "templates_cli",
]



"""Print related methods."""
import click

from statue.cli.styled_strings import name_style, source_style
from statue.evaluation import Evaluation
from statue.verbosity import DEFAULT_VERBOSITY, is_verbose


def title_string(
    original_string: str, underline: str = "=", transform: bool = True
) -> str:
    """
    Create a title string with underline barrier row (markdown style).

    :param original_string: The original string to turn to title
    :type original_string: str
    :param underline: Character to use as underline to the title
    :type underline: str
    :param transform: Transform first letter of each word to upper case
    :type transform: bool
    :return: Markdown title string
    :rtype: str
    """
    unstyled_string = click.unstyle(original_string)
    returned = (
        original_string.replace(unstyled_string, unstyled_string.title())
        if transform
        else original_string
    )
    returned += "\n" + underline * len(unstyled_string)
    return returned


def boxed_string(original_string: str, border: str = "#") -> str:
    """
    String of boxed context.

    :param original_string: The original string to wrap in box
    :type original_string: str
    :param border: Character to use as border to the text
    :type border: str
    :return: Boxed string
    :rtype: str
    """
    unstyled_string = click.unstyle(original_string)
    vertical_border = border * (len(unstyled_string) + 4)
    new_string = original_string.replace(unstyled_string, unstyled_string.title())
    middle_row = f"\n{border} {new_string} {border}\n"
    return vertical_border + middle_row + vertical_border


def evaluation_string(
    evaluation: Evaluation, verbosity: str = DEFAULT_VERBOSITY
) -> str:
    """
    Create evaluation pretty string.

    :param evaluation: The evaluation to format
    :type evaluation: Evaluation
    :param verbosity: Verbosity level of the printing
    :type verbosity: str
    :return: Evaluation as pretty string
    :rtype: str
    """
    returned = ""
    for source, source_evaluation in evaluation.items():
        source_title = title_string(source_style(str(source)), transform=False)
        returned += f"\n\n{source_title}\n\n"
        for command_evaluation in source_evaluation:
            styled_command_name = name_style(command_evaluation.command.name)
            command_title = title_string(
                styled_command_name, underline="-", transform=False
            )
            returned += f"{command_title}\n"
            if is_verbose(verbosity):
                returned += (
                    f"{styled_command_name} ran with args: "
                    f"{command_evaluation.command.args}\n"
                    f"Finished in {command_evaluation.execution_duration:.2f} "
                    "seconds.\n"
                )
            returned += f"{command_evaluation.captured_output_string}\n"
    return returned


def evaluation_summary_string(evaluation: Evaluation) -> str:
    """
    Create a summary string of an evaluation.

    :param evaluation: Evaluation to be printed
    :type evaluation: Evaluation
    :return: summary string
    :rtype: str
    """
    if evaluation.commands_number == 0:
        return "Empty evaluation."
    if evaluation.success:
        return (
            "Statue finished successfully after "
            f"{evaluation.total_execution_duration:.2f} seconds!"
        )
    summary_string = (
        f"Statue has failed after {evaluation.total_execution_duration:.2f} "
        "seconds on the following commands:\n"
    )
    for source, source_evaluation in evaluation.failure_evaluation.items():
        summary_string += f"{source_style(str(source))}:\n"
        summary_string += (
            "\t"
            + ", ".join(
                [
                    name_style(command_evaluation.command.name)
                    for command_evaluation in source_evaluation
                ]
            )
            + "\n"
        )
    return summary_string
